# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Iterable
from copy import deepcopy
from uuid import uuid4

import numpy as np
from compiler_config.config import InlineResultsProcessing

from qat.ir.instructions import (
    Assign,
    BinaryOperator,
    Delay,
    FrequencyShift,
    Instruction,
    InstructionBlock,
    Jump,
    Label,
    PhaseReset,
    PhaseShift,
    Repeat,
    Reset,
    ResultsProcessing,
    Return,
    Synchronize,
)
from qat.ir.measure import (
    Acquire,
    AcquireMode,
    MeasureBlock,
    PostProcessing,
    PostProcessType,
    ProcessAxis,
    acq_mode_process_axis,
)
from qat.ir.pulse_channel import PulseChannel
from qat.ir.waveforms import Pulse, SampledWaveform
from qat.model.device import Component, PhysicalChannel, Qubit
from qat.model.hardware_model import PhysicalHardwareModel
from qat.purr.utils.logger import get_default_logger
from qat.utils.pydantic import QubitId, ValidatedList

log = get_default_logger()


class InstructionBuilder(ABC):
    """Abstract class for assembling quantum programs.

    Provides a number of methods to deal with common quantum semantics, such as qubits,
    gates, measurements, and control flow. The details of how quantum operations are
    implemented are left to child classes.
    """

    def __init__(
        self,
        hardware_model: PhysicalHardwareModel,
        instructions: list[Instruction] = [],
        _qubit_index_by_uuid: dict[str, int] | None = None,
        _qubits_ordered_by_index: list[Qubit] | None = None,
    ):
        self.hw = hardware_model
        self._qubit_index_by_uuid = (
            _qubit_index_by_uuid
            if _qubit_index_by_uuid is not None
            else {qubit.uuid: idx for (idx, qubit) in hardware_model.qubits.items()}
        )
        self._qubits_ordered_by_index = (
            _qubits_ordered_by_index
            if _qubits_ordered_by_index is not None
            else [self.hw.qubits[idx] for idx in sorted(self.hw.qubits.keys())]
        )
        self._ir = InstructionBlock(instructions=instructions)
        self.compiled_shots: int | None = None
        self.shots: int | None = None

    @property
    def qubits(self) -> list[Qubit]:
        """Returns the list of qubits, sorted by index."""
        return self._qubits_ordered_by_index

    def get_physical_qubit(self, index: int) -> Qubit:
        """Returns the qubit assigned with the given physical index.

        :param index: The index of the qubit to return.
        :return: The qubit with the given index.
        """
        return self.hw.qubit_with_index(index)

    def get_logical_qubit(self, index: int) -> Qubit:
        """Returns the qubit assigned with the given logical index.

        :param index: The logical index of the qubit to return.
        :return: The qubit with the given logical index.
        """

        if index >= len(self.qubits):
            raise IndexError(f"Qubit with logical index {index} does not exist.")
        return self.qubits[index]

    @abstractmethod
    def X(
        self, target: Qubit, theta: float = np.pi, pulse_channel: PulseChannel = None
    ): ...

    @abstractmethod
    def Y(
        self, target: Qubit, theta: float = np.pi, pulse_channel: PulseChannel = None
    ): ...

    @abstractmethod
    def Z(
        self, target: Qubit, theta: float = np.pi, pulse_channel: PulseChannel = None
    ): ...

    @abstractmethod
    def U(
        self,
        target: Qubit,
        theta: float,
        phi: float,
        lamb: float,
        pulse_channel: PulseChannel = None,
    ): ...

    @abstractmethod
    def swap(self, target: Qubit, destination: Qubit): ...

    def had(self, target: Qubit):
        return self.Z(target).Y(target, theta=np.pi / 2.0)

    def SX(self, target: Qubit):
        return self.X(target, theta=np.pi / 2.0)

    def SXdg(self, target: Qubit):
        return self.X(target, theta=-(np.pi / 2.0))

    def S(self, target: Qubit):
        return self.Z(target, theta=np.pi / 2.0)

    def Sdg(self, target: Qubit):
        return self.Z(target, theta=-(np.pi / 2))

    def T(self, target: Qubit):
        return self.Z(target, theta=np.pi / 4.0)

    def Tdg(self, target: Qubit):
        return self.Z(target, -(np.pi / 4.0))

    @abstractmethod
    def controlled(self, controllers: Qubit | list[Qubit], builder: InstructionBuilder): ...

    def cX(self, controllers: Qubit | list[Qubit], target: Qubit, theta=np.pi):
        return self.controlled(controllers, self.X(target, theta=theta))

    def cY(self, controllers: Qubit | list[Qubit], target: Qubit, theta=np.pi):
        return self.controlled(controllers, self.Y(target, theta=theta))

    def cZ(self, controllers: Qubit | list[Qubit], target: Qubit, theta=np.pi):
        return self.controlled(controllers, self.Z(target, theta=theta))

    def cnot(self, control: Qubit | list[Qubit], target: Qubit):
        return self.cX(control, target, theta=np.pi)

    @abstractmethod
    def ccnot(self, controllers: list[Qubit], target: Qubit): ...

    def cswap(self, controllers: Qubit | list[Qubit], target: Qubit, destination: Qubit):
        return self.controlled(controllers, self.swap(target, destination))

    @abstractmethod
    def ECR(self, control: Qubit, target: Qubit): ...

    @abstractmethod
    def measure_single_shot_z(
        self,
        target: Qubit,
        axis: ProcessAxis = ProcessAxis.SEQUENCE,
        output_variable: str = None,
    ): ...

    @abstractmethod
    def reset(self, targets: Qubit | list[Qubit], **kwargs): ...

    def repeat(self, repeat_count: int):
        return self.add(Repeat(repeat_count=repeat_count))

    def returns(self, variables: list[str] = None):
        """Add return statement."""
        variables = variables if variables is not None else []
        return self.add(Return(variables=variables))

    def assign(self, name: str, value):
        return self.add(Assign(name=name, value=value))

    def jump(self, label: str | Label, condition: BinaryOperator | None = None):
        return self.add(Jump(label=label, condition=condition))

    def results_processing(self, variable: str, res_format: InlineResultsProcessing):
        return self.add(ResultsProcessing(variable=variable, results_processing=res_format))

    def add(self, *instructions: Instruction, flatten: bool = False):
        """
        Add one or more instruction(s) into this builder. All methods should use this
        instead of accessing the instructions tree directly as it deals with composite instructions.
        """
        self._ir.add(*instructions, flatten=flatten)
        return self

    def __add__(self, other: QuantumInstructionBuilder):
        comp_builder = self.__class__(
            hardware_model=self.hw, instructions=self._ir.instructions.model_copy()
        )
        if isinstance(other, InstructionBuilder):
            comp_builder.add(*other._ir.instructions)
        else:
            raise TypeError(
                "Only another `{self.__class__.__name__}` can be added to this builder."
            )
        return comp_builder

    @staticmethod
    def constrain(angle: float):
        """
        Constrain the rotation angle to avoid redundant rotations around the Bloch sphere.
        """
        return (angle + np.pi) % (2 * np.pi) - np.pi

    @classmethod
    def _check_identity_operation(cls, f):
        """
        Wrapper method to constrain the rotation angle and to determine whether to avoid redundant rotations around the Bloch sphere.
        """

        def wrapper(self, target, theta=np.pi, *args, **kwargs):
            theta = self.constrain(theta)
            return self if np.isclose(theta, 0) else f(self, target, theta, *args, **kwargs)

        return wrapper

    @property
    def instructions(self):
        return self._ir.instructions

    @instructions.setter
    def instructions(self, instructions: ValidatedList[Instruction]):
        self._ir.instructions = instructions

    @property
    def number_of_instructions(self):
        return self._ir.number_of_instructions

    def __iter__(self):
        return self._ir.__iter__()

    def __reversed__(self):
        return self._ir.__reversed__()

    def flatten(self):
        """
        Flatten the instruction builder by removing nested structures like InstructionBlocks.
        """
        self._ir.flatten()
        return self


class QuantumInstructionBuilder(InstructionBuilder):
    """A pulse-level instruction builder, that provides implementations of quantum gates,
    and an API for pulse-level instructions."""

    def __init__(self, *args, **kwargs):
        pulse_channel_map = kwargs.pop("_pulse_channel_map", None)
        super().__init__(*args, **kwargs)
        if pulse_channel_map is None:
            pulse_channel_map = self._build_pulse_channel_mapping(self.hw)
        self._pulse_channels: dict[str, PulseChannel] = pulse_channel_map

    def pretty_print(self):
        output_str = ""
        for instr in self:
            targets = getattr(instr, "targets", None)
            if targets:
                format_target = ""
                for target in targets:
                    format_target = f"{instr.__class__.__name__} -> {self.hw._ids_to_pulse_channels[target]} "
                    device = self.hw._pulse_channel_ids_to_device[target]
                    if isinstance(device, Qubit):
                        format_target += f" @Q{self.hw._qubits_to_qubit_ids[device]}"
                    else:
                        qubit = self.hw._resonators_to_qubits[device]
                        format_target += f" @R{self.hw._qubits_to_qubit_ids[qubit]}"
                output_str += format_target

            else:
                output_str += str(instr)

            output_str += "\n"

        return output_str.rstrip()

    @InstructionBuilder._check_identity_operation
    def X(self, target: Qubit, theta: float = np.pi, pulse_channel: PulseChannel = None):
        """
        Adds a gate that drives the qubit with a rotation angle `theta` to the builder.

        :param target: The qubit to be rotated.
        :param theta: The applied rotation angle.
        :param pulse_channel: The pulse channel the pulses get sent to.
        """
        if np.isclose(np.abs(theta), np.pi / 2.0):
            xpi_pulse = self._hw_X_pi_2(target, pulse_channel=pulse_channel)
            angle = 0 if theta > 0 else np.pi
            return self._apply_z_transform_on_operation(
                target, xpi_pulse, angle, pulse_channel=pulse_channel
            )

        elif (
            np.isclose(np.abs(theta), np.pi)
            and getattr(target, "direct_x_pi", False)
            and getattr(target.drive_pulse_channel, "pulse_x_pi", None) is not None
        ):
            xpi_pulse = self._hw_X_pi(target, pulse_channel=pulse_channel)
            angle = 0 if theta > 0 else np.pi
            return self._apply_z_transform_on_operation(
                target, xpi_pulse, angle, pulse_channel=pulse_channel
            )

        return self.U(
            target,
            theta=theta,
            phi=-np.pi / 2.0,
            lamb=np.pi / 2.0,
            pulse_channel=pulse_channel,
        )

    @InstructionBuilder._check_identity_operation
    def Y(self, target: Qubit, theta: float = np.pi, pulse_channel: PulseChannel = None):
        """
        Adds a gate that drives the qubit with a rotation angle `theta` to the builder.

        :param target: The qubit to be rotated.
        :param theta: The applied rotation angle.
        :param pulse_channel: The pulse channel the pulses get sent to.
        """
        if np.isclose(np.abs(theta), np.pi / 2.0):
            xpi_pulse = self._hw_X_pi_2(target, pulse_channel=pulse_channel)
            angle = np.pi / 2 if theta > 0 else -np.pi / 2
            return self._apply_z_transform_on_operation(
                target, xpi_pulse, angle, pulse_channel=pulse_channel
            )

        elif (
            np.isclose(np.abs(theta), np.pi)
            and getattr(target, "direct_x_pi", False)
            and getattr(target.drive_pulse_channel, "pulse_x_pi", None) is not None
        ):
            xpi_pulse = self._hw_X_pi(target, pulse_channel=pulse_channel)
            angle = np.pi / 2 if theta > 0 else -np.pi / 2
            return self._apply_z_transform_on_operation(
                target, xpi_pulse, angle, pulse_channel=pulse_channel
            )

        return self.U(target, theta=theta, phi=0.0, lamb=0.0, pulse_channel=pulse_channel)

    @InstructionBuilder._check_identity_operation
    def Z(self, target: Qubit, theta: float = np.pi, pulse_channel: PulseChannel = None):
        """
        Adds a virtual gate that rotates the reference frame of the qubit with a phase `theta`
        to the builder.

        :param target: The qubit to be rotated.
        :param theta: The applied rotation angle.
        """
        if pulse_channel:
            log.warning(
                "Pulse channel in Z-gate will be ignored in the `QuantumInstructionBuilder`."
            )

        return self.add(*self._hw_Z(target=target, theta=theta))

    def U(
        self,
        target: Qubit,
        theta: float,
        phi: float,
        lamb: float,
        pulse_channel: PulseChannel = None,
    ):
        """
        Adds an arbitrary rotation around the Bloch sphere with 3 Euler angles
        to the builder (see https://doi.org/10.48550/arXiv.1707.03429).

        :param target: The qubit to be rotated.
        :param theta: Rotation angle.
        :param phi: Rotation angle.
        :param lamb: Rotation angle.
        :param pulse_channel: The pulse channel the pulses get sent to.
        """
        theta = self.constrain(theta)
        return self.add(
            *self._hw_Z(target, theta=lamb + np.pi, pulse_channel=pulse_channel),
            *self._hw_X_pi_2(target, pulse_channel=pulse_channel),
            *self._hw_Z(target, theta=np.pi - theta, pulse_channel=pulse_channel),
            *self._hw_X_pi_2(target, pulse_channel=pulse_channel),
            *self._hw_Z(target, theta=phi, pulse_channel=pulse_channel),
        )

    def ZX(self, target1: Qubit, target2: Qubit, theta: float = np.pi / 4.0):
        """
        Adds a two-qubit interaction gate exp(-i \theta Z x X) to the builder.

        :param target1: The qubit to which Z gets applied to.
        :param target2: The qubit to which X gets applied to.
        :param theta: The applied rotation angle.
        """
        theta = self.constrain(theta)
        if np.isclose(theta, 0):
            return []

        target1_id = self._qubit_index_by_uuid[target1.uuid]
        target2_id = self._qubit_index_by_uuid[target2.uuid]

        cr_pulse_channel = target1.cross_resonance_pulse_channels[target2_id]
        cr_cancellation_pulse_channel = target2.cross_resonance_cancellation_pulse_channels[
            target1_id
        ]

        if np.isclose(theta, np.pi / 4.0):
            return self.add(*self._hw_ZX_pi_4(target1, target2))
        elif np.isclose(theta, -np.pi / 4.0):
            return self.add(
                PhaseShift(targets=cr_pulse_channel.uuid, phase=np.pi),
                PhaseShift(targets=cr_cancellation_pulse_channel.uuid, phase=np.pi),
                *self._hw_ZX_pi_4(target1, target2),
                PhaseShift(targets=cr_pulse_channel.uuid, phase=np.pi),
                PhaseShift(targets=cr_cancellation_pulse_channel.uuid, phase=np.pi),
            )
        else:
            raise NotImplementedError(
                "Generic ZX gate not implemented yet! Please use `theta` = pi/4 or -pi/4."
            )

    def _apply_z_transform_on_operation(
        self,
        target: Qubit,
        gate: list[Pulse],
        theta: float,
        pulse_channel: PulseChannel = None,
    ):
        if np.isclose(theta, 0.0):
            return self.add(*gate)

        return self.add(
            *self._hw_Z(target, theta=-theta, pulse_channel=pulse_channel),
            *gate,
            *self._hw_Z(target, theta=theta, pulse_channel=pulse_channel),
        )

    def _hw_X_pi_2(
        self, target: Qubit, pulse_channel: PulseChannel = None, amp_scale: float = 1.0
    ):
        """Op definition for a X(pi/2) gate.

        Optionally allows a pulse channel to be provided to apply the X(pi/2) pulse down.
        """

        pulse_channel = pulse_channel or self.get_pulse_channel(
            target.drive_pulse_channel.uuid
        )
        pulse_waveform = target.drive_pulse_channel.pulse
        pulse_waveform = pulse_waveform.waveform_type(**pulse_waveform.model_dump())
        pulse_waveform.amp *= amp_scale
        return [Pulse(targets=pulse_channel.uuid, waveform=pulse_waveform)]

    def _hw_X_pi(self, target: Qubit, pulse_channel: PulseChannel = None):
        pulse_channel = pulse_channel or self.get_pulse_channel(
            target.drive_pulse_channel.uuid
        )
        pulse_waveform = getattr(target.drive_pulse_channel, "pulse_x_pi", None)
        if pulse_waveform is None or not getattr(target, "direct_x_pi", False):
            x_pi_2_pulse = deepcopy(self._hw_X_pi_2(target, pulse_channel=pulse_channel))
            return [*x_pi_2_pulse, *x_pi_2_pulse]
        pulse_waveform = pulse_waveform.waveform_type(**pulse_waveform.model_dump())
        return [Pulse(targets=pulse_channel.uuid, waveform=pulse_waveform)]

    def _hw_Z(
        self, target: Qubit, theta: float = np.pi, pulse_channel: PulseChannel = None
    ):
        """Op definition for a Z(theta) gate.

        If a channel is provided, the phase shift is applied to that channel. If none is
        provided, or the channel is the drive channel for the target qubit, the phase shift
        is applied to all pulse channels associated with the qubit.
        """

        if theta == 0:
            return []

        # Rotate drive pulse channel of the qubit.
        qubit_drive_id = target.drive_pulse_channel.uuid
        pulse_channel = pulse_channel or self.get_pulse_channel(qubit_drive_id)
        instr_collection = [PhaseShift(targets=pulse_channel.uuid, phase=theta)]
        # Rotate all cross resonance (cancellation) pulse channels pertaining to the qubit.
        qubit_id = self._qubit_index_by_uuid[target.uuid]
        if pulse_channel.uuid == qubit_drive_id:
            for (
                coupled_qubit_id,
                crc_pulse_channel,
            ) in target.cross_resonance_cancellation_pulse_channels.items():
                coupled_qubit = self.hw.qubit_with_index(coupled_qubit_id)
                cr_pulse_channel = coupled_qubit.cross_resonance_pulse_channels[qubit_id]

                instr_collection.append(
                    PhaseShift(targets=cr_pulse_channel.uuid, phase=theta)
                )
                instr_collection.append(
                    PhaseShift(targets=crc_pulse_channel.uuid, phase=theta)
                )

        return instr_collection

    def _hw_ZX_pi_4(self, target1: Qubit, target2: Qubit):
        """Native op definition for a ZX(pi/4) gate."""

        target1_id = self._qubit_index_by_uuid[target1.uuid]
        target2_id = self._qubit_index_by_uuid[target2.uuid]

        target1_pulse_channel = target1.cross_resonance_pulse_channels[target2_id]
        target2_pulse_channel = target2.cross_resonance_cancellation_pulse_channels[
            target1_id
        ]

        if target1_pulse_channel is None or target2_pulse_channel is None:
            raise ValueError(
                f"Tried to perform cross resonance on {str(target2)} "
                f"that is not linked to {str(target1)}."
            )

        pulse = target1_pulse_channel.zx_pi_4_pulse.model_dump()
        waveform_type = target1_pulse_channel.zx_pi_4_pulse.waveform_type
        if pulse is None:
            raise ValueError(
                f"No `zx_pi_4_pulse` available on {target1} with index {target1_id}."
            )

        sync = Synchronize(
            targets=[
                pulse_ch.uuid
                for pulse_ch in target1.all_qubit_and_resonator_pulse_channels
                + target2.all_qubit_and_resonator_pulse_channels
            ]
        )

        return [
            sync,
            Pulse(target=target1_pulse_channel.uuid, waveform=waveform_type(**pulse)),
            Pulse(target=target2_pulse_channel.uuid, waveform=waveform_type(**pulse)),
            sync,
        ]

    def _generate_measure_acquire(
        self, target: Qubit, mode: AcquireMode, output_variable: str = None
    ):
        # Measure-related info.
        measure_channel = target.measure_pulse_channel
        measure_instruction = Pulse(
            targets=measure_channel.uuid,
            waveform=measure_channel.pulse.waveform_type(
                **measure_channel.pulse.model_dump()
            ),
        )

        # Acquire-related info.
        acquire_channel = target.resonator.acquire_pulse_channel
        acquire_duration = (
            measure_channel.pulse.width
            if acquire_channel.acquire.sync
            else acquire_channel.acquire.width
        )

        if acquire_channel.acquire.use_weights:
            filter = Pulse(
                waveform=SampledWaveform(samples=acquire_channel.acquire.weights),
                duration=acquire_duration,
                target=acquire_channel.uuid,
            )
        else:
            filter = None

        A, B = target.mean_z_map_args[0], target.mean_z_map_args[1]
        mean_g = (1 - B) / A
        mean_e = (-1 - B) / A
        rotation = np.mod(-np.angle(mean_e - mean_g), 2 * np.pi)
        threshold = (np.exp(1j * rotation) * (mean_e + mean_g)).real / 2

        acquire_instruction = Acquire(
            targets=acquire_channel.uuid,
            duration=acquire_duration,
            mode=mode,
            delay=acquire_channel.acquire.delay,
            filter=filter,
            output_variable=output_variable,
            rotation=rotation,
            threshold=threshold,
        )

        return [
            measure_instruction,
            acquire_instruction,
        ]

    def measure(
        self,
        targets: Qubit | set[Qubit],
        mode: AcquireMode = AcquireMode.INTEGRATOR,
        output_variable: str = None,
        sync_qubits: bool = True,
    ) -> QuantumInstructionBuilder:
        """
        Measure one or more qubits.

        :param targets: The qubit(s) to be measured.
        :param mode: The type of acquisition at the level of the control hardware.
        :param sync_qubits: Flag determining whether to align the measurements of all
                            qubits in `targets` or not. Sync between qubits is on by default.
        """
        if isinstance(targets, Qubit):
            targets = [targets]

        target_ids = {self._qubit_index_by_uuid[target.uuid] for target in targets}
        measure_block = MeasureBlock(qubit_targets=target_ids)

        all_pulse_channels = [
            pc.uuid
            for qubit in targets
            for pc in qubit.all_qubit_and_resonator_pulse_channels
        ]

        if sync_qubits:
            # Sync pulse channels of all pulse channels.
            measure_block.add(Synchronize(targets=all_pulse_channels))

        for qubit in targets:
            if not sync_qubits:
                qubit_pulse_channels = [
                    pc.uuid for pc in qubit.all_qubit_and_resonator_pulse_channels
                ]
                measure_block.add(Synchronize(targets=qubit_pulse_channels))

            output_var = output_variable or self._generate_output_variable(qubit)

            # Measure and acquire instructions for the qubit.
            measure, acquire = self._generate_measure_acquire(
                qubit, mode=mode, output_variable=output_var
            )
            duration = max(measure.duration, acquire.delay + acquire.duration)
            measure_block.add(measure, acquire)
            measure_block.duration = max(measure_block.duration, duration)

            if not sync_qubits:
                qubit_pulse_channels = [
                    pc.uuid for pc in qubit.all_qubit_and_resonator_pulse_channels
                ]
                measure_block.add(Synchronize(targets=qubit_pulse_channels))

        # Sync pulse channels of all pulse channels after measurement.
        if sync_qubits:
            measure_block.add(Synchronize(targets=all_pulse_channels))

        return self.add(measure_block)

    def _take_time_mean(
        self,
        acquire_mode: AcquireMode,
        target: Qubit | set[Qubit],
        output_variable: str,
    ):
        """
        Takes the mean of the acquired data.
        If the acquire mode is INTEGRATOR, it returns the builder without any changes

        :param target: The qubit to be measured.
        :param axis: The type of axis which the post-processing of readouts should occur on.
        :param output_variable: The variable where the output of the acquire should be saved.
        """
        if acquire_mode == AcquireMode.INTEGRATOR:
            return self

        return self.post_processing(
            target, output_variable, PostProcessType.MEAN, ProcessAxis.TIME
        )

    def _take_sequence_mean(
        self,
        acquire_mode: AcquireMode,
        target: Qubit | set[Qubit],
        output_variable: str,
    ):
        """
        Applies the mean post-processing to the acquired data along the sequence axis. If the acquire mode is SCOPE,
        the mean post-processing is not needed, so it returns the builder without any changes.

        :param target: The qubit to be measured.
        :param axis: The type of axis which the post-processing of readouts should occur on.
        :param output_variable: The variable where the output of the acquire should be saved.
        """
        if acquire_mode == AcquireMode.SCOPE:
            return self

        return self.post_processing(
            target, output_variable, PostProcessType.MEAN, ProcessAxis.SEQUENCE
        )

    def measure_single_shot_z(
        self,
        target: Qubit,
        axis: ProcessAxis = ProcessAxis.SEQUENCE,
        output_variable: str = None,
    ):
        """
        Measure a single qubit along the z-axis.
        Execution results in an array of floats centred around 0, where anything above 0 shows a bias towards a
        +Z measurement, and below 0 shows a bias towards a -Z measurement.
        The array is an ndarray, where the first dimension is for each of the qubits measured, the second dimension is
        for each shot run.

        :param target: The qubit to be measured.
        :param axis: The type of axis which the post-processing of readouts should occur on.
        :param output_variable: The variable where the output of the acquire should be saved.
        """
        output_variable = output_variable or self._generate_output_variable(target)
        acquire_mode = acq_mode_process_axis[axis]
        self.measure(target, acquire_mode, output_variable)
        self._take_time_mean(acquire_mode, target, output_variable)
        return self.post_processing(
            target, output_variable, PostProcessType.LINEAR_MAP_COMPLEX_TO_REAL
        )

    def measure_single_shot_signal(
        self,
        target: Qubit,
        axis: ProcessAxis = ProcessAxis.SEQUENCE,
        output_variable: str = None,
    ):
        """
        Measure the signal of a single qubit.
        Execution results in an array of complex numbers that show the war signal output form the hardware.
        The array is an ndarray, where the first dimension is for each of the qubits measured, the second dimension is
        for each shot run.

        :param target: The qubit to be measured.
        :param axis: The type of axis which the post-processing of readouts should occur on.
        :param output_variable: The variable where the output of the acquire should be saved.
        """
        output_variable = output_variable or self._generate_output_variable(target)
        acquire_mode = acq_mode_process_axis[axis]
        self.measure(target, acquire_mode, output_variable)
        return self._take_time_mean(acquire_mode, target, output_variable)

    def measure_mean_z(
        self,
        target: Qubit,
        axis: ProcessAxis = ProcessAxis.SEQUENCE,
        output_variable: str = None,
    ):
        """
        Measure a single qubit along the z-axis.
        Execution results in an array of floats centred around 0, where anything above 0 shows a bias towards a
        +Z measurement, and below 0 shows a bias towards a -Z measurement.
        Each entry is for each qubit measured, and is the mean of the results from `measure_single_shot_z`.

        :param target: The qubit to be measured.
        :param axis: The type of axis which the post-processing of readouts should occur on.
        :param output_variable: The variable where the output of the acquire should be saved.
        """
        output_variable = output_variable or self._generate_output_variable(target)
        acquire_mode = acq_mode_process_axis[axis]
        self.measure(target, acquire_mode, output_variable)
        self._take_time_mean(acquire_mode, target, output_variable)
        self._take_sequence_mean(acquire_mode, target, output_variable)
        return self.post_processing(
            target, output_variable, PostProcessType.LINEAR_MAP_COMPLEX_TO_REAL
        )

    def measure_mean_signal(self, target: Qubit, output_variable: str = None):
        """
        Measure the signal of a single qubit.
        Execution results in an array of complex numbers that show the war signal output form the hardware.
        Each entry is for each qubit measured, and is the mean of the results from `measure_single_signal_shot`.

        :param target: The qubit to be measured.
        :param output_variable: The variable where the output of the acquire should be saved.
        """
        output_variable = output_variable or self._generate_output_variable(target)
        acquire_mode = acq_mode_process_axis[ProcessAxis.SEQUENCE]
        return self.measure(target, acquire_mode, output_variable)._take_sequence_mean(
            acquire_mode, target, output_variable
        )

    def measure_scope_mode(self, target: Qubit, output_variable: str = None):
        """
        Measure the scope mode

        :param target: The qubit to be measured.
        :param output_variable: The variable where the output of the acquire should be saved.
        """
        acquire_mode = acq_mode_process_axis[ProcessAxis.TIME]
        return self.measure(target, acquire_mode, output_variable)

    def measure_single_shot_binned(
        self,
        target: Qubit,
        axis: ProcessAxis = ProcessAxis.SEQUENCE,
        output_variable: str = None,
    ):
        """
        Measure a single qubit along the z-axis.
        Execution results in an array of ±1s, where 1 indicates a +Z measurement, and -1 indicates a -Z measurement.
        It takes the results from `measure_single_shot_z` and applies a binning operation to round the results to ±1.
        The array is an ndarray, where the first dimension is for each of the qubits measured, the second dimension is
        for each shot run.

        :param target: The qubit to be measured.
        :param axis: The type of axis which the post-processing of readouts should occur on.
        :param output_variable: The variable where the output of the acquire should be saved.
        """
        output_variable = output_variable or self._generate_output_variable(target)
        acquire_mode = acq_mode_process_axis[axis]
        return (
            self.measure(target, acq_mode_process_axis[axis], output_variable)
            ._take_time_mean(acquire_mode, target, output_variable)
            .post_processing(
                target, output_variable, PostProcessType.LINEAR_MAP_COMPLEX_TO_REAL
            )
            .post_processing(target, output_variable, PostProcessType.DISCRIMINATE)
        )

    def reset(
        self, targets: PulseChannel | Qubit | Iterable[PulseChannel | Qubit], **kwargs
    ):
        if isinstance(targets, list):
            targets = set(targets)
        elif isinstance(targets, PulseChannel | Qubit):
            targets = {targets}
        else:
            raise TypeError(
                f"Invalid type, expected '(PulseChannel | Qubit | list[PulseChannel | Qubit])' but got {type(targets)}."
            )

        pulse_channel_ids = []
        for target in targets:
            if isinstance(target, Qubit):
                self.add(Reset(qubit_target=self._qubit_index_by_uuid[target.uuid]))
                pulse_channel_ids.extend(
                    [pc.uuid for pc in target.all_qubit_and_resonator_pulse_channels]
                )
            else:
                pulse_channel_ids.append(target.uuid)

        for pulse_ch_id in pulse_channel_ids:
            self.add(PhaseReset(target=pulse_ch_id))

        return self

    def _generate_output_variable(self, target: Component):
        return "out_" + target.uuid + f"_{np.random.randint(np.iinfo(np.int32).max)}"

    def _find_valid_measure_block(self, target_ids: QubitId | set[QubitId]):
        """
        Finds the previous :class:`MeasureBlock` where the given qubits are not measured yet.

        :param target_ids: The indices of the qubits to be measured.
        """
        for instruction in reversed(self.instructions):
            if isinstance(instruction, MeasureBlock) and not any(
                target_ids & instruction.qubit_targets
            ):
                instruction.qubit_targets.update(target_ids)
                return instruction
        return None

    def post_processing(
        self,
        target: Qubit,
        output_variable: str,
        process_type: PostProcessType,
        axes: ProcessAxis | list[ProcessAxis] | None = None,
        args=None,
    ):
        axes = axes if axes is not None else []
        args = args if args is not None else []

        # Default the mean z-map args if none supplied.
        if not any(args):
            if process_type == PostProcessType.LINEAR_MAP_COMPLEX_TO_REAL:
                args = target.mean_z_map_args

            elif process_type == PostProcessType.DISCRIMINATE:
                args = [target.discriminator]

        axes = axes if isinstance(axes, list) else [axes]

        return self.add(
            PostProcessing(
                output_variable=output_variable,
                process_type=process_type,
                axes=axes,
                args=args,
            )
        )

    def cnot(self, control: Qubit, target: Qubit):
        return (
            self.ECR(control, target)
            .X(control)
            .Z(control, theta=-np.pi / 2)
            .X(target, theta=-np.pi / 2)
        )

    def ECR(self, control: Qubit, target: Qubit):
        if not isinstance(control, Qubit) or not isinstance(target, Qubit):
            raise ValueError("The quantum targets of the ECR node must be qubits!")

        target_id = self._qubit_index_by_uuid[target.uuid]

        if not control.cross_resonance_pulse_channels.get(target_id, None):
            raise ValueError(
                f"Qubits {self._qubit_index_by_uuid[control.uuid]} and {target_id} are not coupled."
            )

        return (
            self.ZX(control, target, theta=np.pi / 4.0)
            .X(control, theta=np.pi)
            .ZX(control, target, theta=-np.pi / 4.0)
        )

    def swap(self, target: Qubit, destination: Qubit):
        raise NotImplementedError("Not available on this hardware model.")

    def pulse(self, **kwargs):
        return self.add(Pulse(**kwargs))

    def acquire(
        self,
        target: Qubit,
        delay: float = 1e-06,
        **kwargs,
    ):
        pulse_channel = target.acquire_pulse_channel

        if delay is None:
            kwargs["delay"] = pulse_channel.acquire.delay

        return self.add(Acquire(target=pulse_channel.uuid, **kwargs))

    def delay(self, target: Qubit | PulseChannel, duration: float):
        delays = []
        if isinstance(target, Qubit):
            delays = [
                Delay(targets=pulse_ch.uuid, duration=duration)
                for pulse_ch in target.all_qubit_and_resonator_pulse_channels
            ]
        else:
            delays = [Delay(targets=target.uuid, duration=duration)]
        return self.add(*delays)

    def synchronize(self, targets: Qubit | list[Qubit | PulseChannel]):
        targets = targets if isinstance(targets, list) else [targets]

        pulse_channel_ids = set()
        for target in targets:
            if isinstance(target, Qubit):
                qubit_pulse_channel_ids = [
                    pulse_channel.uuid
                    for pulse_channel in target.all_qubit_and_resonator_pulse_channels
                ]
                pulse_channel_ids.update(qubit_pulse_channel_ids)
            else:
                pulse_channel_ids.add(target.uuid)
        if len(pulse_channel_ids) > 1:
            return self.add(Synchronize(targets=pulse_channel_ids))
        else:
            log.warning(
                f"Ignored synchronisation of a single pulse channel with id {next(iter(pulse_channel_ids))}."
            )
            return self

    def controlled(self, controllers: Qubit | list[Qubit], builder: InstructionBuilder):
        raise NotImplementedError("Not available on this hardware model.")

    def ccnot(self, controllers: list[Qubit], target: Qubit):
        raise NotImplementedError("Not available on this hardware model.")

    @InstructionBuilder._check_identity_operation
    def phase_shift(self, target: PulseChannel, theta: float):
        return self.add(PhaseShift(targets=target.uuid, phase=theta))

    def frequency_shift(self, target: PulseChannel, frequency: float):
        if np.isclose(frequency, 0.0):
            return self
        return self.add(FrequencyShift(targets=target.uuid, frequency=frequency))

    def get_pulse_channel(
        self,
        id: str,
    ) -> PulseChannel:
        """Given an id, return the corresponding pulse channel.

        Checks internally stored pulse channels, but can pull pulse channels from the
        hardware model if not found.
        """

        pulse_channel = self._pulse_channels.get(id, None)
        if pulse_channel is not None:
            return pulse_channel

        pulse_channel = self.hw.pulse_channel_with_id(id)
        if pulse_channel is None:
            raise ValueError(f"Pulse channel with id '{id}' not found.")

        pulse_channel = self.create_pulse_channel(
            frequency=pulse_channel.frequency,
            physical_channel=self.hw.physical_channel_for_pulse_channel_id(
                pulse_channel.uuid
            ).uuid,
            imbalance=pulse_channel.imbalance,
            phase_iq_offset=pulse_channel.phase_iq_offset,
            scale=pulse_channel.scale,
            uuid=pulse_channel.uuid,
        )
        return pulse_channel

    def create_pulse_channel(
        self,
        frequency: float,
        physical_channel: PhysicalChannel | str,
        imbalance: float = 1.0,
        phase_iq_offset: float = 0.0,
        scale: float | complex = 1.0 + 0.0j,
        uuid: str | None = None,
    ) -> PulseChannel:
        """Creates a pulse channel and adds stores it within the builder.

        The channel can be provided as a physical channel which the logical channel is
        linked too, or use the physical channel of a provided pulse channel.
        """

        if isinstance(physical_channel, PhysicalChannel):
            physical_channel = physical_channel.uuid
        else:
            if self.hw.physical_channel_with_id(physical_channel) is None:
                raise ValueError(
                    f"Physical channel with id '{physical_channel}' not found."
                )

        uuid = uuid if uuid is not None else str(uuid4())
        pulse_channel = PulseChannel(
            uuid=uuid,
            physical_channel_id=physical_channel,
            frequency=frequency,
            imbalance=imbalance,
            phase_iq_offset=phase_iq_offset,
            scale=scale,
        )

        self._pulse_channels[uuid] = pulse_channel
        return pulse_channel

    @staticmethod
    def _build_pulse_channel_mapping(hw: PhysicalHardwareModel) -> dict[str, PulseChannel]:
        pulse_channels = {}
        for qubit in hw.qubits.values():
            for device in [qubit, qubit.resonator]:
                physical_channel_id = device.physical_channel.uuid
                for pc in device.all_pulse_channels:
                    pulse_channels[pc.uuid] = PulseChannel(
                        physical_channel_id=physical_channel_id,
                        frequency=pc.frequency,
                        imbalance=pc.imbalance,
                        phase_iq_offset=pc.phase_iq_offset,
                        scale=pc.scale,
                        uuid=pc.uuid,
                    )
        return pulse_channels

    def _create_empty_builder(self):
        """Creates a new instance of the builder that is free of instructions; passes
        through the expensive mappings created at instantiation."""

        return type(self)(
            hardware_model=self.hw,
            instructions=[],
            _pulse_channel_map=self._pulse_channels,
            _qubit_index_by_uuid=self._qubit_index_by_uuid,
            _qubits_ordered_by_index=self._qubits_ordered_by_index,
        )


PydQuantumInstructionBuilder = QuantumInstructionBuilder



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from pydantic import BaseModel, Field


class QRPairModel(BaseModel):
    """
    Used to parse JSON/CSV entries. Qubit and Resonator are specified together in pairs.
    """

    label: str = Field(alias="LABEL")
    index: int = Field(frozen=True, alias="INDEX")
    connections: str = Field(alias="CONNECTIONS")

    q_freq: float = Field(alias="Q-FREQ")
    q_freq12: float = Field(alias="Q-FREQ12")
    q_cluster: str = Field(alias="Q-CLUSTER")
    q_slot_idx: int = Field(alias="Q-SLOT-IDX")
    q_output: int = Field(alias="Q-OUTPUT")

    r_freq: float = Field(alias="R-FREQ")
    r_cluster: str = Field(alias="R-CLUSTER")
    r_slot_idx: int = Field(alias="R-SLOT-IDX")
    r_output: int = Field(alias="R-OUTPUT")
    r_input: int = Field(alias="R-INPUT")

    def qubit_physical_channel_id(self):
        return f"{self.q_cluster}-QCM-RF-{self.q_slot_idx}-OUT-{self.q_output}"

    def resonator_physical_channel_id(self):
        return f"{self.r_cluster}-QRM-RF-{self.r_slot_idx}-OUT-{self.r_output}-IN-{self.r_input}"



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd
from __future__ import annotations

from typing import Annotated

import numpy as np
from pydantic import BeforeValidator, Field, PrivateAttr, field_validator, model_validator

from qat.ir.instructions import (
    Delay,
    Instruction,
    QuantumInstruction,
    QuantumInstructionBlock,
    Synchronize,
)
from qat.ir.waveforms import Pulse

# The following things from legacy instructions are unchanged, so just import for now.
from qat.purr.compiler.instructions import AcquireMode, PostProcessType, ProcessAxis
from qat.utils.pydantic import QubitId, ValidatedList, ValidatedSet, _validate_set


class Acquire(QuantumInstruction):
    suffix_incrementor: int = 0
    duration: float = 1e-6
    mode: AcquireMode = AcquireMode.INTEGRATOR
    delay: float | None = 0.0
    filter: Pulse | None = Field(default=None)
    output_variable: str | None = None
    rotation: float | None = 0.0
    threshold: float | None = 0.0

    @property
    def pulse_channel(self):
        return next(iter(self.targets))

    @property
    def target(self):
        return next(iter(self.targets))

    @model_validator(mode="before")
    def _validate_filter(cls, data: dict):
        if isinstance(data, dict):
            duration = data.get("duration")
            filter = data.get("filter")

            if filter:
                filter_duration = (
                    filter.duration if isinstance(filter, Pulse) else filter["duration"]
                )
                # TODO: COMPILER-722 -- with this in place, it required to add a duration to both the pulse
                #  and the Acquire instruction, would be better to figure out how to share information in some cases.
                if filter_duration == 0:  # < 0 condition already tested in `Waveform`
                    raise ValueError("Filter duration cannot be equal to zero.")

                if not np.isclose(filter_duration, duration, atol=1e-9):
                    raise ValueError(
                        f"Filter duration '{filter_duration}' must be equal to Acquire "
                        f"duration '{duration}'."
                    )

        return data


class PostProcessing(Instruction):
    """
    States what post-processing should happen after data has been acquired. This can
    happen in the FPGA's or a software post-process.
    """

    output_variable: str | None = None
    process_type: PostProcessType
    axes: list[ProcessAxis] = []
    args: list[float | complex] = []
    result_needed: bool = False

    @classmethod
    def _from_legacy(cls, legacy_pp):
        # private as we dont want to support this in the long-term
        return cls(
            output_variable=legacy_pp.output_variable,
            process_type=legacy_pp.process.value,
            axes=legacy_pp.axes,
            args=legacy_pp.args,
            result_needed=legacy_pp.result_needed,
        )

    @field_validator("axes", mode="before")
    @classmethod
    def _axes_as_list(cls, axes=None):
        if axes:
            return axes if isinstance(axes, list) else [axes]
        return []

    @field_validator("args", mode="before")
    def _validate_args(cls, args=[]):
        """Ensures that the args are not numpy arrays or numpy numbers."""

        args = [args] if not isinstance(args, (list, np.ndarray)) else args
        return np.asarray(args).tolist()


VALID_MEASURE_INSTR = Synchronize | Acquire | Pulse | Delay


class MeasureBlock(QuantumInstructionBlock):
    """
    Encapsulates a measurement of a single (or multiple) qubit(s).
    It should only contain instructions that are associated with a
    measurement such as a measure pulse, an acquire or a synchronize.
    """

    instructions: ValidatedList[VALID_MEASURE_INSTR] = Field(
        default_factory=lambda: ValidatedList[VALID_MEASURE_INSTR]()
    )

    qubit_targets: Annotated[ValidatedSet[QubitId], BeforeValidator(_validate_set)] = Field(
        default_factory=lambda: ValidatedSet()
    )
    _output_variables: list[str] = PrivateAttr(default=[])

    def add(self, *instructions: QuantumInstruction):
        for instruction in instructions:
            if isinstance(instruction, Acquire) and (delay := instruction.delay) > 0.0:
                self._output_variables.append(instruction.output_variable)
                super().add(Delay(target=instruction.target, duration=delay))
                instruction.delay = 0.0
            super().add(instruction)
        return self

    @property
    def output_variables(self):
        return self._output_variables


acq_mode_process_axis = {
    ProcessAxis.SEQUENCE: AcquireMode.INTEGRATOR,
    ProcessAxis.TIME: AcquireMode.SCOPE,
    "sequence": AcquireMode.INTEGRATOR,
    "time": AcquireMode.SCOPE,
}



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd

import numpy as np
from matplotlib import pyplot as plt

from qat.backend.qblox.acquisition import Acquisition
from qat.backend.qblox.execution import QbloxProgram
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


def plot_program(program: QbloxProgram):
    packages = program.packages

    if not packages:
        return

    max_length = max([len(pkg.timeline) for pkg in packages.values()])
    if max_length <= 0:
        return

    # Padding short timelines with zeros
    for pkg in packages.values():
        length = len(pkg.timeline)
        if length < max_length:
            pkg.timeline = np.append(
                pkg.timeline, np.zeros(max_length - length, dtype=pkg.timeline.dtype)
            )

    t = np.linspace(0, max_length, max_length)
    fig, axes = plt.subplots(
        nrows=len(packages),
        ncols=1,
        sharex=False,
        sharey=False,
        squeeze=False,
        figsize=(10, 5),
    )
    fig.suptitle("Timeline plots")
    for i, pkg in enumerate(packages.values()):
        axes[i][0].plot(t, pkg.timeline.real, label="I")
        axes[i][0].plot(t, pkg.timeline.imag, label="Q")
        axes[i][0].set_title(pkg.pulse_channel_id)
        axes[i][0].set_xlabel("Time (ns)")
        axes[i][0].set_ylabel("Digital offset")
        axes[i][0].autoscale()
        axes[i][0].legend()

    plt.tight_layout()
    plt.show()


def plot_playback(playback: dict[str, list[Acquisition]]):
    if not playback:
        return

    for i, (pulse_channel_id, acquisitions) in enumerate(playback.items()):
        for acquisition in acquisitions:
            fig, axes = plt.subplots(
                nrows=3,
                ncols=1,
                sharex=False,
                sharey=False,
                squeeze=False,
                figsize=(10, 5),
            )
            fig.suptitle(f"Playback plots for {acquisition.name} on {pulse_channel_id}")

            scope_data = acquisition.acquisition.scope
            integ_data = acquisition.acquisition.bins.integration
            thrld_data = acquisition.acquisition.bins.threshold

            # Scope data
            axes[0, 0].plot(scope_data.path0.data, label="I")
            axes[0, 0].plot(scope_data.path1.data, label="Q")
            axes[0, 0].set_xlabel("Sample (ns)")
            axes[0, 0].set_ylabel("Value")
            axes[0, 0].autoscale()
            axes[0, 0].legend()
            axes[0, 0].title.set_text("Scope acquisition")

            # Integration data
            axes[1, 0].plot(integ_data.path0, label="I")
            axes[1, 0].plot(integ_data.path1, label="Q")
            axes[1, 0].set_xlabel("Iteration")
            axes[1, 0].set_ylabel("Value")
            axes[1, 0].autoscale()
            axes[1, 0].legend()
            axes[1, 0].title.set_text("Integrated acquisition")

            # Threshold data
            axes[2, 0].plot(thrld_data, label="I")
            axes[2, 0].set_xlabel("Iteration")
            axes[2, 0].set_ylabel("Value")
            axes[2, 0].autoscale()
            axes[2, 0].title.set_text("Thresholded acquisition")

        plt.tight_layout()
        plt.show()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from qat.core.pass_base import TransformPass
from qat.purr.compiler.builders import InstructionBuilder
from qat.purr.compiler.devices import PulseChannel, PulseChannelView
from qat.purr.compiler.instructions import DeviceUpdate, FrequencySet


class FrequencyAssignSanitisation(TransformPass):
    """Sanitises device assignments that changes the frequency of a channel, by replacing
    it with a FrequencySet instruction.

    If this is done for targets that can support frequency assignment as an IR instruction,
    (i.e. has support for :class:`FrequencySet`), then this pass can be used to avoid
    mutating the hardware model, which can be unsafe when used within pipelines.
    """

    def run(self, ir: InstructionBuilder, *args, **kwargs):
        instructions = []
        for instruction in ir._instructions:
            if (
                isinstance(instruction, DeviceUpdate)
                and isinstance(instruction.target, (PulseChannel, PulseChannelView))
                and instruction.attribute == "frequency"
            ):
                instructions.append(
                    FrequencySet(channel=instruction.target, frequency=instruction.value)
                )
            else:
                instructions.append(instruction)
        ir._instructions = instructions
        return ir



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
import inspect
from pathlib import Path

import piny
from pydantic import ImportString, ValidationInfo, field_validator, model_validator
from pydantic_settings import SettingsConfigDict

from qat.core.config.descriptions import (
    CompilePipelineDescription,
    EngineDescription,
    ExecutePipelineDescription,
    HardwareLoaderDescription,
    PipelineClassDescription,
    PipelineFactoryDescription,
    PipelineInstanceDescription,
    UpdateablePipelineDescription,
)
from qat.extensions import QatExtension
from qat.purr.qatconfig import QatConfig
from qat.utils.piny import VeryStrictMatcher


class QatSessionConfig(QatConfig):
    """
    Configuration for QAT() sessions extending qat.purr.qatconfig configuration.

    >>> import os
    >>> os.environ["QAT_MAX_REPEATS_LIMIT"] = "654321"
    >>> QatSessionConfig() # doctest: +ELLIPSIS
    QatSessionConfig(MAX_REPEATS_LIMIT=654321, ...)
    >>> QatSessionConfig(MAX_REPEATS_LIMIT=123) # doctest: +ELLIPSIS
    QatSessionConfig(MAX_REPEATS_LIMIT=123, ...)
    >>> del os.environ["QAT_MAX_REPEATS_LIMIT"]
    >>> qatconfig = QatSessionConfig()
    >>> qatconfig.MAX_REPEATS_LIMIT = 16000
    >>> qatconfig # doctest: +ELLIPSIS
    QatSessionConfig(MAX_REPEATS_LIMIT=16000, ...)

    >>> QatSessionConfig(MAX_REPEATS_LIMIT=100.5) # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    pydantic_core._pydantic_core.ValidationError
    ...
    Input should be a valid integer, got a number with a fractional part
    """

    model_config = SettingsConfigDict(
        env_prefix="QAT_",
        env_nested_delimiter="_",
        validate_assignment=True,
        yaml_file="qatconfig.yaml",
    )

    EXTENSIONS: list[ImportString] = []
    """ QAT Extensions to initialise on start-up"""

    HARDWARE: list[HardwareLoaderDescription] = []
    """ QAT Hardware Models to load on start-up """

    ENGINES: list[EngineDescription] = []
    """ QAT Engines to add on start-up """

    PIPELINES: (
        list[
            PipelineInstanceDescription
            | UpdateablePipelineDescription
            | PipelineFactoryDescription
            | PipelineClassDescription
        ]
        | None
    ) = None
    """ QAT Pipelines for compilation and execution to add on start-up, None adds default
    pipelines"""

    COMPILE: (
        list[
            PipelineInstanceDescription
            | UpdateablePipelineDescription
            | PipelineFactoryDescription
            | CompilePipelineDescription
        ]
        | None
    ) = None
    """ QAT Pipelines for compilation to add on start-up, None adds default pipelines"""

    EXECUTE: (
        list[
            PipelineInstanceDescription
            | UpdateablePipelineDescription
            | PipelineFactoryDescription
            | ExecutePipelineDescription
        ]
        | None
    ) = None
    """ QAT Pipelines for execution to add on start-up, None adds default pipelines"""

    @field_validator("EXTENSIONS")
    def load_extensions(values):
        for value in values:
            if type(value) is type and issubclass(value, QatExtension):
                value.load()
            else:
                name = inspect.getmodule(value).__name__
                raise ValueError(
                    f"extension '{name}' must be a valid QatExtension class, type is "
                    f"actually {type(value)}"
                )
        return values

    @model_validator(mode="after")
    def one_default(cls, values: dict) -> dict:
        """Ensures that exactly one PIPELINE is marked as default, or alternatively, one of
        COMPILE and EXECUTE pipelines is marked as default."""
        compile_pipelines = values.COMPILE
        execute_pipelines = values.EXECUTE
        full_pipelines = values.PIPELINES

        if not (compile_pipelines or execute_pipelines):
            if full_pipelines and len(full_pipelines) == 1:
                full_pipelines[0].default = True
                return values

        num_full_defaults = (
            sum(desc.default for desc in full_pipelines)
            if full_pipelines is not None
            else 0
        )
        if num_full_defaults > 1:
            raise ValueError(
                "Expected at most one default pipeline in PIPELINES, found "
                f"{num_full_defaults}."
            )

        for pipelines in [compile_pipelines, execute_pipelines]:
            if not pipelines and not full_pipelines:
                continue

            num_pipelines = len(pipelines) if pipelines is not None else 0
            num_pipeline_defaults = (
                sum(desc.default for desc in pipelines) if num_pipelines != 0 else 0
            )

            if num_pipelines == 1 and not full_pipelines:
                pipelines[0].default = True
                continue

            if (total_defaults := num_pipeline_defaults + num_full_defaults) != 1:
                raise ValueError(
                    f"Expected exactly one default COMPILE and EXECUTE pipelines, found "
                    f"multiple (including PIPELINES): {total_defaults}."
                )

        return values

    @field_validator("HARDWARE", "ENGINES")
    @classmethod
    def no_duplicate_names(cls, val, info: ValidationInfo):
        """Checks that there are no duplicate names in the list of descriptions."""
        if val is None or len(val) == 0:
            return val

        names = set()
        for desc in val:
            if desc.name in names:
                raise ValueError(f"Duplicate name {desc.name} found in {info.field_name}")
            names.add(desc.name)
        return val

    @field_validator("COMPILE", "EXECUTE")
    @classmethod
    def no_duplicate_pipeline_names(cls, val, info: ValidationInfo):
        """For each of the PIPELINES, COMPILE, and EXECUTE fields, this checks that there
        are no duplicate names in the list of descriptions. Also checks there are no shared
        names between PIPELINES and COMPILE/EXECUTE."""
        from qat.pipelines.pipeline import CompilePipeline, ExecutePipeline

        pipeline_ty = CompilePipeline if info.field_name == "COMPILE" else ExecutePipeline

        pipelines = [] if val is None else val
        full_pipelines = info.data.get("PIPELINES", [])
        full_pipelines = [] if full_pipelines is None else full_pipelines
        full_pipelines = [
            desc for desc in full_pipelines if desc.is_subtype_of(pipeline_ty)
        ]

        names = set()
        for desc in full_pipelines + pipelines:
            if desc.name in names:
                raise ValueError(
                    f"Duplicate name {desc.name} found in {info.field_name} pipelines."
                )
            names.add(desc.name)
        return val

    @field_validator("PIPELINES", "ENGINES", "COMPILE", "EXECUTE")
    @classmethod
    def matching_hardware_loaders(cls, pipelines, info: ValidationInfo):
        if pipelines is None:
            return pipelines
        if len(pipelines) == 0:
            return pipelines

        hardware_loader_names = {hld.name for hld in info.data.get("HARDWARE", [])}
        for Pdesc in pipelines:
            expected_loader = getattr(Pdesc, "hardware_loader", None)
            if expected_loader is not None and expected_loader not in hardware_loader_names:
                raise ValueError(f"Hardware Loader {expected_loader} not defined")

        return pipelines

    @field_validator("PIPELINES", "EXECUTE")
    @classmethod
    def matching_engines(cls, pipelines, info: ValidationInfo):
        if pipelines is None:
            return pipelines
        if len(pipelines) == 0:
            return pipelines

        engine_names = {desc.name for desc in info.data.get("ENGINES", [])}
        engine_loaders = {
            desc.name: desc.hardware_loader for desc in info.data.get("ENGINES", [])
        }

        for Pdesc in pipelines:
            engine = getattr(Pdesc, "engine", None)
            if engine is None:
                continue

            if engine not in engine_names:
                raise ValueError(
                    f"Pipeline {Pdesc.name} requires engine {engine}, but it is not "
                    "defined in ENGINES"
                )

            expected_loader = getattr(Pdesc, "hardware_loader", None)
            engine_loader = engine_loaders[engine]
            if (
                engine_loader is not None
                and expected_loader is not None
                and engine_loader != expected_loader
            ):
                raise ValueError(
                    f"Pipeline {Pdesc.name} has hardware_loader {expected_loader}, but "
                    f"the engine {engine} has a different hardware_loader {engine_loader}."
                )

        return pipelines

    @classmethod
    def from_yaml(cls, path: str | Path):
        blob = piny.YamlLoader(path=str(path), matcher=VeryStrictMatcher).load()
        return cls(**blob)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from qat.model.loaders.purr import RTCSModelLoader
from qat.pipelines.legacy.base import LegacyExecutePipeline as LegacyRTCSExecutePipeline
from qat.pipelines.updateable import PipelineConfig

from .compile import LegacyRTCSCompilePipeline
from .full import LegacyRTCSPipeline

legacy_rtcs2 = LegacyRTCSPipeline(
    config=PipelineConfig(name="legacy_rtcs2"),
    loader=RTCSModelLoader(),
).pipeline

__all__ = [
    "LegacyRTCSPipeline",
    "LegacyRTCSCompilePipeline",
    "LegacyRTCSExecutePipeline",
    "PipelineConfig",
    "legacy_rtcs2",
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

import csv
import os

from qat.purr.backends.qblox.execution.executor import CompositeExecutor, LeafExecutor
from qat.purr.backends.qblox.execution.instrument_base import InstrumentModel


def load_executor(instrument_info_csv: str):
    """
    Builds a ControlHardware object wrapping an arbitrary fleet of Qblox clusters defined as CSV
    """

    if not os.path.exists(instrument_info_csv):
        raise ValueError(f"File '{instrument_info_csv}' not found!")

    executor = CompositeExecutor()
    with open(instrument_info_csv) as f:
        reader = csv.DictReader(f)
        for row in reader:
            instrument = InstrumentModel.model_validate(row)
            executor.add(LeafExecutor(instrument))

    return executor



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

import re

import numpy as np
from compiler_config.config import CompilerConfig

from qat.core.pass_base import TransformPass
from qat.purr.backends.qiskit_simulator import QiskitBuilder
from qat.purr.compiler.error_mitigation.readout_mitigation import get_readout_mitigation
from qat.purr.compiler.instructions import Assign, is_generated_name


class QiskitErrorMitigation(TransformPass):
    """Implements readout error mitigation for legacy Qiskit engines.

    Because the legacy Qiskit engine returns results in a format that is not consistent with
    other legacy engines (or refactored ones), it requires its own implementation. The
    refactored engine will be written so that the output of the results are formatted
    in a way that the runtime will expect to receive them. This is just a work around to
    support legacy engines in the new pipelines API.
    """

    # the pattern assign variables should follow
    index_pattern = re.compile(r"(.*)\[(?P<clbit_index>[0-9]+)\]_(?P<qubit_index>[0-9]+)")

    def run(
        self,
        acquisitions: dict[str, any],
        *args,
        compiler_config: CompilerConfig,
        package: QiskitBuilder,
        **kwargs,
    ) -> dict[str, any]:
        """
        :param acquisitions: Acquisition data returned from the Qiskit simulator.
        :param compiler_config: The compiler config contains the error mitigation
            configuration.
        :param package: The Qiskit instruction builder.
        :raises ValueError: Multiple registers are not allowed with error mitigation.
        :return: The transformed result acquisitions.
        """
        error_mitigation = compiler_config.error_mitigation
        if error_mitigation is None:
            return acquisitions

        if len(acquisitions) > 1:
            raise ValueError(
                "Cannot have multiple registers in conjunction with readout error "
                "mitigation."
            )

        mapping = self.classical_to_quantum_mapping(package)
        for mitigator in get_readout_mitigation(error_mitigation):
            new_result = mitigator.apply_error_mitigation(acquisitions, mapping, self.model)
            acquisitions[mitigator.name] = new_result
        return acquisitions

    def classical_to_quantum_mapping(self, package: QiskitBuilder) -> dict[str, int]:
        """Generates a mapping between classical register indices and qubit indices.

        This would probably be an analysis pass for refactored code, but since its only used
        here, its just included within this transformation pass.

        :param package: The Qiskit instruction builder.
        :raises ValueError: If the assign variables is not in the expected format, an error
            is thrown.
        :return: Returns the mapping as a dictionary.
        """
        mapping = {}
        for instruction in package.instructions:
            if not isinstance(instruction, Assign):
                continue
            for value in instruction.value:
                result = self.index_pattern.match(value.name)
                if result is None:
                    raise ValueError(
                        "Could not extract cl register index from "
                        f"{instruction.output_variable}."
                    )
                mapping[result.group("clbit_index")] = int(result.group("qubit_index"))
        return mapping


class QiskitStripMetadata(TransformPass):
    """Detects if Qiskit is returning meta data associated with the simulation, and trims
    it away if so."""

    def run(self, acquisitions: dict[str, any], *args, **kwargs):
        """
        :param acquisitions: Acquisition data returned from the Qiskit simulator.
        :return: The acquisition data with metadata stripped away (if found).
        """
        if isinstance(acquisitions, tuple):
            return acquisitions[0]
        else:
            return acquisitions


class QiskitSimplifyResults(TransformPass):
    """Strips away randomly generated names.

    This is the same as the `simplify_results` function in :class:`ResultsTransform`.
    """

    def run(self, acquisitions: dict[str, any], *args, **kwargs):
        """
        :param acquisitions: Acquisition data returned from the Qiskit simulator.
        :return: The acquisition data simplified.
        """
        if all([is_generated_name(k) for k in acquisitions.keys()]):
            if len(acquisitions) == 1:
                acquisitions = list(acquisitions.values())[0]
            else:
                squashed_results = list(acquisitions.values())
                if all(isinstance(val, np.ndarray) for val in squashed_results):
                    acquisitions = np.array(squashed_results)
                else:
                    acquisitions = squashed_results
        return acquisitions



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from qat.model.loaders.purr.echo import EchoModelLoader
from qat.pipelines.updateable import PipelineConfig

from .compile import LegacyEchoCompilePipeline
from .execute import LegacyEchoExecutePipeline
from .full import LegacyEchoPipeline


def _create_pipeline_instance(num_qubits: int) -> LegacyEchoPipeline:
    return LegacyEchoPipeline(
        config=PipelineConfig(name=f"legacy_echo{num_qubits}"),
        loader=EchoModelLoader(qubit_count=num_qubits),
    ).pipeline


legacy_echo8 = _create_pipeline_instance(8)
legacy_echo16 = _create_pipeline_instance(16)
legacy_echo32 = _create_pipeline_instance(32)

__all__ = [
    "LegacyEchoPipeline",
    "LegacyEchoCompilePipeline",
    "LegacyEchoExecutePipeline",
    "PipelineConfig",
    "legacy_echo8",
    "legacy_echo16",
    "legacy_echo32",
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
import abc


class QatExtension:
    """Seperately installed QatExtensions can be registered in qatconfig and will be loaded automatically by calling the load method

    These are used internally at OQC for proprietary extensions.
    """

    @staticmethod
    @abc.abstractmethod
    def load():
        pass



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from qat.model.loaders.purr.base import BaseLegacyModelLoader
from qat.purr.backends.echo import Connectivity, get_default_echo_hardware
from qat.purr.compiler.hardware_models import QuantumHardwareModel


class EchoModelLoader(BaseLegacyModelLoader):
    def __init__(
        self,
        qubit_count: int = 4,
        connectivity: Connectivity | list[(int, int)] | None = None,
        add_direction_couplings=True,
    ):
        self.connectivity = connectivity
        self.qubit_count = qubit_count
        self.add_direction_couplings = add_direction_couplings

    def load(self) -> QuantumHardwareModel:
        return get_default_echo_hardware(
            qubit_count=self.qubit_count,
            connectivity=self.connectivity,
            add_direction_couplings=self.add_direction_couplings,
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

import abc

from compiler_config.config import CompilerConfig

from qat.core.metrics_base import MetricsManager
from qat.core.pass_base import PassManager
from qat.core.result_base import ResultManager
from qat.purr.compiler.hardware_models import QuantumHardwareModel


class BaseMiddleend(abc.ABC):
    """
    Base class for a middle end that takes an intermediate representation (IR) :class:`QatIR`
    and alters it based on optimisation and/or validation passes.
    """

    def __init__(self, model: None | QuantumHardwareModel):
        """
        :param model: The hardware model that holds calibrated information on the qubits on the QPU.
        """
        self.model = model

    @abc.abstractmethod
    def emit(
        self,
        ir,
        res_mgr: ResultManager | None = None,
        met_mgr: MetricsManager | None = None,
        compiler_config: CompilerConfig | None = None,
        **kwargs,
    ):
        """
        Converts an IR :class:`QatIR` to an optimised IR.
        :param ir: The intermediate representation.
        :param res_mgr: Collection of analysis results with caching and aggregation
            capabilities, defaults to None.
        :param met_mgr: Stores useful intermediary metrics that are generated during
            compilation, defaults to None.
        :param compiler_config: Compiler settings, defaults to None.
        """
        ...


class CustomMiddleend(BaseMiddleend):
    """
    Middle end that uses a custom pipeline to convert the IR to an (optimised) IR.
    """

    def __init__(
        self, model: None | QuantumHardwareModel, pipeline: None | PassManager = None
    ):
        self.pipeline = pipeline
        super().__init__(model=model)

    def emit(
        self,
        ir,
        res_mgr: ResultManager | None = None,
        met_mgr: MetricsManager | None = None,
        compiler_config: CompilerConfig | None = None,
        **kwargs,
    ):
        """
        Converts an IR :class:`QatIR` to an optimised IR with a custom pipeline.
        :param ir: The intermediate representation.
        :param res_mgr: Collection of analysis results with caching and aggregation
            capabilities, defaults to None.
        :param met_mgr: Stores useful intermediary metrics that are generated during
            compilation, defaults to None.
        :param compiler_config: Compiler settings, defaults to None.
        """

        res_mgr = res_mgr if res_mgr is not None else ResultManager()
        met_mgr = met_mgr if met_mgr is not None else MetricsManager()
        compiler_config = (
            compiler_config if compiler_config is not None else CompilerConfig()
        )

        ir = self.pipeline.run(
            ir, res_mgr, met_mgr, compiler_config=compiler_config, **kwargs
        )
        return ir



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd

import json
import os
from collections import defaultdict
from dataclasses import asdict
from datetime import datetime
from typing import Dict, List

from qblox_instruments import Cluster
from qblox_instruments.qcodes_drivers.module import Module
from qblox_instruments.qcodes_drivers.sequencer import Sequencer

from qat.purr.backends.live_devices import ControlHardware, Instrument, LivePhysicalBaseband
from qat.purr.backends.qblox.acquisition import Acquisition
from qat.purr.backends.qblox.codegen import QbloxPackage
from qat.purr.backends.qblox.config import (
    QbloxConfig,
    QcmConfigHelper,
    QcmRfConfigHelper,
    QrmConfigHelper,
    QrmRfConfigHelper,
)
from qat.purr.backends.qblox.constants import Constants
from qat.purr.backends.qblox.visualisation import plot_packages, plot_playback
from qat.purr.compiler.devices import (
    ChannelType,
    PhysicalChannel,
    PulseChannel,
    Qubit,
    Resonator,
)
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class QbloxResonator(Resonator):
    """
    A hack around QBlox acquisition to use the same pulse channel
    for MeasurePulse and Acquire instructions.
    """

    def get_measure_channel(self) -> PulseChannel:
        return self.get_pulse_channel(ChannelType.macq)

    def get_acquire_channel(self) -> PulseChannel:
        return self.get_pulse_channel(ChannelType.macq)


class QbloxPhysicalBaseband(LivePhysicalBaseband):
    def __init__(
        self,
        id_,
        frequency,
        if_frequency,
        instrument: Instrument,
        slot_idx: int,
        config: QbloxConfig,
    ):
        super().__init__(id_, frequency, if_frequency, instrument)
        self.slot_idx = slot_idx
        self.config = config  # TBD, is it worth flattening ?


class QbloxPhysicalChannel(PhysicalChannel):
    baseband: QbloxPhysicalBaseband

    @property
    def slot_idx(self):
        return self.baseband.slot_idx

    @property
    def config(self):
        return self.baseband.config

    def build_resonator(self, resonator_id, *args, **kwargs):
        """Helper method to build a resonator with default channels."""
        kwargs.pop("fixed_if", None)
        reson = QbloxResonator(resonator_id, self)
        reson.create_pulse_channel(ChannelType.macq, *args, fixed_if=False, **kwargs)
        reson.default_pulse_channel_type = ChannelType.macq
        return reson

    def build_qubit(
        self,
        index,
        resonator,
        drive_freq,
        second_state_freq=None,
        channel_scale=(1.0e-8 + 0.0j),
        measure_amp: float = 1.0,
        fixed_drive_if=False,
        qubit_id=None,
    ):
        """
        Helper method tp build a qubit with assumed default values on the channels. Modelled after the live hardware.
        """
        qubit = Qubit(index, resonator, self, drive_amp=measure_amp, id_=qubit_id)
        qubit.create_pulse_channel(
            ChannelType.drive,
            frequency=drive_freq,
            scale=channel_scale,
            fixed_if=False,
        )

        qubit.create_pulse_channel(
            ChannelType.second_state, frequency=second_state_freq, scale=channel_scale
        )

        qubit.measure_acquire["delay"] = 151e-9  # TOF

        return qubit


class QbloxControlHardware(ControlHardware):
    _driver: Cluster

    def __init__(
        self,
        dev_id: str,
        name: str,
        address: str = None,
        dummy_cfg: Dict = None,
    ):
        super().__init__(id_=dev_id)
        self.name = name
        self.address = address
        self.dummy_cfg = dummy_cfg

        self.dump_sequence = False
        self.plot_packages = False
        self.plot_acquisitions = False

        self._modules: Dict[Module, bool] = {}
        self._id2seq: Dict[str, Sequencer] = {}

    def _reset_modules(self):
        # TODO - Qblox bug: Hard reset clutters sequencer connections with conflicting defaults
        # TODO - This is a temporary workaround until Qblox fixes the issue

        modules = [mod for mod, is_dirty in self._modules.items() if is_dirty]

        for mod in modules:
            log.debug(f"Resetting sequencer connections for module {mod}")
            mod.disconnect_outputs()
            if mod.is_qrm_type:
                mod.disconnect_inputs()
            self._modules[mod] = False  # Mark as clean

    @property
    def id2seq(self):
        return self._id2seq

    def configure(self, package: QbloxPackage):
        module: Module = getattr(self._driver, f"module{package.slot_idx}")
        sequencer = module.sequencers[package.seq_idx]

        seq_config, mod_config = package.seq_config, package.mod_config
        if module.is_qcm_type:
            if module.is_rf_type:
                config_helper = QcmRfConfigHelper(mod_config, seq_config)
            else:
                config_helper = QcmConfigHelper(mod_config, seq_config)
        elif module.is_qrm_type:
            if module.is_rf_type:
                config_helper = QrmRfConfigHelper(mod_config, seq_config)
            else:
                config_helper = QrmConfigHelper(mod_config, seq_config)
        else:
            raise ValueError(f"Unknown module type {module.module_type}")

        try:
            log.debug(f"Configuring module {module}, and sequencer {sequencer}")
            config_helper.configure(module, sequencer)
            sequence = asdict(package.sequence)
            sequencer.sequence(sequence)
        finally:
            self._modules[module] = True  # Mark as dirty
            self._id2seq[package.pulse_channel_id] = sequencer

    def connect(self):
        if self._driver is None or not Cluster.is_valid(self._driver):
            self._driver: Cluster = Cluster(
                name=self.name,
                identifier=self.address,
                dummy_cfg=self.dummy_cfg if self.address is None else None,
            )
            self._driver.reset()
            self._modules = {m: True for m in self._driver.get_connected_modules().values()}
            self.is_connected = True

        log.info(self._driver.get_system_status())

    def disconnect(self):
        if self._driver is not None:
            try:
                self._driver.close()
                self._driver = None
                self._modules.clear()
                self.is_connected = False
            except BaseException as e:
                log.warning(
                    f"Failed to close instrument ID: {self.id} at: {self.address}\n{str(e)}"
                )

    def set_data(self, packages: List[QbloxPackage]):
        if self.plot_packages:
            plot_packages(packages)

        if self.dump_sequence:
            for pkg in packages:
                filename = f"schedules/target_{pkg.pulse_channel_id}_@_{datetime.now().strftime('%m-%d-%Y_%H%M%S')}.json"
                os.makedirs(os.path.dirname(filename), exist_ok=True)
                with open(filename, "w") as f:
                    f.write(json.dumps(asdict(pkg.sequence)))

        try:
            self._id2seq.clear()
            self._reset_modules()
            for pkg in packages:
                self.configure(pkg)
        except BaseException as e:
            self._id2seq.clear()
            self._reset_modules()
            raise e

    def start_playback(self, repetitions: int, repetition_time: float):
        if not any(self._id2seq):
            raise ValueError("No allocations found. Install packages and configure first")

        results: Dict[str, List[Acquisition]] = defaultdict(list)
        try:
            for pulse_channel_id, sequencer in self._id2seq.items():
                sequencer.sync_en(True)

            for pulse_channel_id, sequencer in self._id2seq.items():
                sequencer.arm_sequencer()

            for pulse_channel_id, sequencer in self._id2seq.items():
                sequencer.start_sequencer()

            for pulse_channel_id, sequencer in self._id2seq.items():
                if ChannelType.macq.name in pulse_channel_id:
                    status_obj = sequencer.get_sequencer_status()
                    log.debug(f"Sequencer status - {sequencer}: {status_obj}")
                    if acquisitions := sequencer.get_acquisitions():
                        # TODO - 60 min tops, make it dynamic by involving flow-aware timeline duration
                        # Only wait if you sequencer is expected to have acquisitions
                        # TODO - Precise expectation of acquisitions should come from higher up
                        sequencer.get_acquisition_status(timeout=60)

                    for name in acquisitions:
                        sequencer.store_scope_acquisition(name)

                    # (re)fetch the lot after having stored scope acquisition
                    acquisitions = sequencer.get_acquisitions()

                    integ_length = sequencer.integration_length_acq()
                    start, end = (
                        0,
                        min(
                            integ_length,
                            Constants.MAX_SAMPLE_SIZE_SCOPE_ACQUISITIONS,
                        ),
                    )
                    for name, acquisition in acquisitions.items():
                        acquisition = Acquisition.model_validate(acquisition)
                        acquisition.name = name

                        scope_data = acquisition.acquisition.scope
                        scope_data.path0.data = scope_data.path0.data[start:end]
                        scope_data.path1.data = scope_data.path1.data[start:end]

                        integ_data = acquisition.acquisition.bins.integration
                        integ_data.path0 /= integ_length
                        integ_data.path1 /= integ_length

                        results[pulse_channel_id].append(acquisition)

            if self.plot_acquisitions:
                plot_playback(results)

            return results
        finally:
            for pulse_channel_id, sequencer in self._id2seq.items():
                sequencer.sync_en(False)
                if ChannelType.macq.name in pulse_channel_id:
                    sequencer.delete_acquisition_data(all=True)

            self._id2seq.clear()
            self._reset_modules()

    def __getstate__(self) -> Dict:
        results = super(QbloxControlHardware, self).__getstate__()
        results["_driver"] = None
        results["_modules"] = {}
        results["_id2seq"] = {}
        return results

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._driver = None
        self._modules = {}
        self._id2seq = {}
        self.is_connected = False



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from abc import ABC, abstractmethod

import numpy as np
from pydantic import NonNegativeInt

from qat.ir.gates.base import Angle, GateBase


class Gate2Q(GateBase, ABC):
    """Base gate for two qubit operations."""

    qubit1: NonNegativeInt
    qubit2: NonNegativeInt

    @property
    def qubits(self):
        return (self.qubit1, self.qubit2)

    @property
    @abstractmethod
    def matrix(self):
        pass


class ControlGate2Q(Gate2Q, ABC):
    r"""Base class for controlled two-qubit gates. The convention is :attr:`qubit1` is the
    control qubit and :attr:`qubit` is the target qubit. For some single qubit gate
    :math:`G`, the controlled gate is

    .. math:: CG = \ket{1}\bra{1} \otimes G + \ket{0}\bra{0} \otimes I

    Will eventually be replaced by a mechanism for creating control gates from standard
    gates. Used now to mark gates as control gates (which will be considered in future
    optimisation passes).
    """

    @property
    def control(self):
        return self.qubit1

    @property
    def target(self):
        return self.qubit2


class CNOT(ControlGate2Q):
    r"""Implements a CNOT with the control gate on :param:`qubit1` and target gate on
    :param:`qubit2`.
    
    Matrix representation:
    
    .. math:: {\rm CNOT}_{c, t} = \begin{bmatrix}
        1 & 0 & 0 & 0 \\
        0 & 1 & 0 & 0 \\
        0 & 0 & 0 & 1 \\
        0 & 0 & 1 & 0
        \end{bmatrix}
    """

    @property
    def matrix(self):
        return np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]])


CX = CNOT


class ECR(Gate2Q):
    r"""Implements the ECR gate.
    
    Matrix representation:
    
    .. math:: {\rm ECR} = \begin{bmatrix}
        0 & 0 & 1 & i \\
        0 & 0 & i & 1 \\
        1 & -i & 0 & 0 \\
        -i & 1 & 0 & 0
        \end{bmatrix}
    """

    @property
    def matrix(self):
        return (1 / np.sqrt(2)) * np.array(
            [[0, 0, 1, 1j], [0, 0, 1j, 1], [1, -1j, 0, 0], [-1j, 1, 0, 0]]
        )


class SWAP(Gate2Q):
    r"""Implements the SWAP gate.
    
    Matrix representation:
    
    .. math:: {\rm SWAP} = \begin{bmatrix}
        1 & 0 & 0 & 0 \\
        0 & 0 & 1 & 0 \\
        0 & 1 & 0 & 0 \\
        0 & 0 & 0 & 1
        \end{bmatrix}
    """

    @property
    def matrix(self):
        return np.array([[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]])


class CPhase(ControlGate2Q):
    r"""Implements the controlled-:math:`Phase` gate.

    Unlike the non-controlled gate, note that this is not equivalent to the
    :math:`CR_{z}(\theta)` gate.
    """

    theta: Angle = np.pi

    @property
    def matrix(self):
        return np.array(
            [
                [1, 0, 0, 0],
                [0, 1, 0, 0],
                [0, 0, 1, 0],
                [0, 0, 0, np.exp(1j * self.theta)],
            ]
        )


class CRz(ControlGate2Q):
    r"""Implements the controlled-:math:`R_{z}(\theta)` gate."""

    theta: Angle = np.pi

    @property
    def matrix(self):
        return np.array(
            [
                [1, 0, 0, 0],
                [0, 1, 0, 0],
                [0, 0, np.exp(-1j * self.theta / 2), 0],
                [0, 0, 0, np.exp(1j * self.theta / 2)],
            ]
        )


class CRx(ControlGate2Q):
    r"""Implements the controlled-:math:`R_{x}(\theta)` gate."""

    theta: Angle = np.pi

    @property
    def matrix(self):
        cos = np.cos(self.theta / 2)
        sin = np.sin(self.theta / 2)
        return np.array(
            [
                [1, 0, 0, 0],
                [0, 1, 0, 0],
                [0, 0, cos, -1j * sin],
                [0, 0, -1j * sin, cos],
            ]
        )


class CRy(ControlGate2Q):
    r"""Implements the controlled-:math:`R_{y}(\theta)` gate."""

    theta: Angle = np.pi

    @property
    def matrix(self):
        cos = np.cos(self.theta / 2)
        sin = np.sin(self.theta / 2)
        return np.array(
            [
                [1, 0, 0, 0],
                [0, 1, 0, 0],
                [0, 0, cos, -sin],
                [0, 0, sin, cos],
            ]
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd
import itertools

import numpy as np

from qat.backend.passes.purr.analysis import BindingResult, TriageResult
from qat.core.metrics_base import MetricsManager
from qat.core.pass_base import TransformPass
from qat.core.result_base import PreservedResults, ResultManager
from qat.purr.compiler.builders import InstructionBuilder
from qat.purr.compiler.instructions import Instruction, Sweep


class DesugaringPass(TransformPass):
    """Transforms syntactic sugars and implicit effects.

    The goal of this pass is to desugar any syntactic and semantic sugars. One of these
    sugars is iteration constructs such as :class:`Sweep` and :class:`Repeat`.

    1- Adds an iterator variable for sweep loops. We state that sweep loops implicitly
       declare an iterator variable to control progress of the sweep. See :class:`BindingPass`
    """

    def run(self, ir: InstructionBuilder, res_mgr: ResultManager, *args, **kwargs):
        """:param ir: The list of instructions stored in an :class:`InstructionBuilder`."""

        for inst in ir.instructions:
            if isinstance(inst, Sweep):
                count = len(next(iter(inst.variables.values())))
                iter_name = f"{hash(inst)}"
                inst.variables[iter_name] = np.linspace(1, count, count, dtype=int).tolist()
        return ir


class ScopePeeling(TransformPass):
    """
    A transform pass to discard scopes assuming they’ve been identified.

    This pass is particularly useful for the code generator as it helps peel away any unwanted scopes.
    In fact, some sweeps are not lowerable and the user may choose to annotate certain sweeps and opt
    for static injection of variables with a switerator-like emitter. These sweep loops will be handled
    by the emitter statically while allowing it to lower nested constructs efficiently.
    """

    def run(
        self,
        ir: InstructionBuilder,
        res_mgr: ResultManager,
        met_mgr: MetricsManager,
        *args,
        **kwargs,
    ):
        scopes: list[tuple[Instruction, Instruction]] = kwargs.get("scopes", [])

        if scopes is None or len(scopes) == 0:
            return PreservedResults.all()

        if any(not isinstance(scope, tuple) or len(scope) != 2 for scope in scopes):
            raise ValueError(
                f"Invalid scopes argument. Expected a list of tuples of length 2, got {scopes}"
            )

        if any(s is None or e is None for (s, e) in scopes):
            raise ValueError(
                "Invalid scopes argument. Delimiter instructions must not be none."
            )

        # Results assumed to be correct at this level of analysis
        triage_result = res_mgr.lookup_by_type(TriageResult)
        binding_result = res_mgr.lookup_by_type(BindingResult)

        for target, instructions in triage_result.target_map.items():
            scoping_result = binding_result.scoping_results[target]
            for scope in scopes:
                if scope not in scoping_result.scope2symbols:
                    raise ValueError(
                        f"Could not find scope {scope}, the binding analysis result may have been corrupt"
                    )

        delimiters = set(itertools.chain(*scopes))

        def predicate(inst):
            return inst not in delimiters

        instructions = ir.splice()
        ir.instructions = list(filter(predicate, instructions))

        return PreservedResults.discard(triage_result, binding_result)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from numbers import Number

import numpy as np

from qat.ir.measure import PostProcessing
from qat.purr.compiler.instructions import AcquireMode, PostProcessType, ProcessAxis

# TODO: move to QAT config?
UPCONVERT_SIGN = 1.0


def get_axis_map(mode: AcquireMode, response: np.ndarray) -> dict[ProcessAxis, int]:
    """
    Given the acquisition mode, determine what each axis corresponds to.

    The way the results are returned are defined by the acquisition mode: this could be
    averaged over shots, averaged over time, or neither. We must determine how to unpack
    the results.

    :param mode: The acquisition mode.
    :param response: The response returned by the target machine.
    :returns: A dictionary containing the axis index for each type of :class:`ProcessAxis`.
    """

    match mode:
        case AcquireMode.SCOPE:
            return {ProcessAxis.TIME: -1}
        case AcquireMode.INTEGRATOR:
            return {ProcessAxis.SEQUENCE: -1}
        case AcquireMode.RAW:
            if response.ndim > 1:
                return {ProcessAxis.SEQUENCE: -2, ProcessAxis.TIME: -1}
            else:
                return {ProcessAxis.TIME: -1}
    raise NotImplementedError(f"The acquisition mode {mode} is not supported.")


def apply_post_processing(
    response: np.ndarray, post_processing: PostProcessing, axes: dict[ProcessAxis, int]
) -> tuple[np.ndarray, dict[ProcessAxis, int]]:
    """
    Applies software post processing to the results.

    Uses the information in the :class:`PostProcessing` instruction to determine what method
    to  apply.

    :param response: Readout results from an execution engine.
    :param post_processing: The post processing instruction.
    :param axes: A dictionary containing which axes contain the shots and which contain time
        series.
    :returns: The processed results as an array and the axis map.
    """

    match post_processing.process_type:
        case PostProcessType.MEAN:
            return mean(response, axes, post_processing.axes)
        case PostProcessType.LINEAR_MAP_COMPLEX_TO_REAL:
            return linear_map_complex_to_real(response, axes, *post_processing.args)
        case PostProcessType.DISCRIMINATE:
            return discriminate(response, axes, post_processing.args[0])

    raise NotImplementedError(
        f"Post processing type {post_processing.process_type} not implemented."
    )


def mean(
    response: np.ndarray,
    axes: dict[ProcessAxis, int],
    target_axes: ProcessAxis | list[ProcessAxis],
):
    """
    Calculates the mean over the given axes.

    :param response: Readout results from an execution engine.
    :param axes: A dictionary containing which axes contain the shots and which contain time
        series.
    :param target_axes: Which axis or axes should the mean be done over?
    :returns: The processed results as an array and the axis map.
    """

    target_axes = target_axes if isinstance(target_axes, list) else [target_axes]
    axis_indices = tuple(axes[axis] for axis in axes if axis in target_axes)
    final_axes = _remove_axes(response.ndim, axis_indices, axes)
    final_data = np.mean(response, axis=axis_indices)
    return final_data, final_axes


def linear_map_complex_to_real(
    response: list[np.ndarray],
    axes: dict[ProcessAxis, int],
    multiplier: Number,
    constant: Number,
):
    """
    Maps complex values onto a real z-projection using a provided linear mapping.

    :param np.ndarray response: Readout results from an execution engine.
    :param axes: A dictionary containing which axes contain the shots and which contain time
        series.
    :type axes: dict[ProcessAxis, Int]
    :param numbers.Number multiplier: Coeffecient for the linear map.
    :param numbers.Number constant: Constant for the linear map.
    :returns: The processed results as an array and the axis map.
    """

    return np.real(multiplier * response + constant), axes


def discriminate(response: np.ndarray, axes: dict[ProcessAxis, int], threshold: float):
    """
    Discriminates a real value to a classical bit by comparison to a supplied
    discrimination threshold.

    :param np.ndarray response: Readout results from an execution engine.
    :param axes: A dictionary containing which axes contain the shots and which contain time
        series.
    :type axes: dict[ProcessAxis, Int]
    :param float threshold: The supplied discrimination threshold.
    :returns: The processed results as an array and the axis map.
    """
    return 2 * (response > threshold) - 1, axes


def _remove_axes(original_dims, removed_axis_indices, axis_locations):
    """
    Extracted from `purr/backends/utilities.py`.

    Returns the new axis map after axes have been removed, e.g., from calculating a mean.
    """

    # map original axis index to new axis index
    axis_map = {i: i for i in range(original_dims)}
    for r in removed_axis_indices:
        if r < 0:
            r = original_dims + r
        axis_map[r] = None
        for i in range(r + 1, original_dims):
            if axis_map[i] is not None:
                axis_map[i] -= 1
    axis_negative = {k: v < 0 for k, v in axis_locations.items()}
    new_axis_locations = {
        k: (original_dims + v if v < 0 else v) for k, v in axis_locations.items()
    }
    new_axis_locations = {
        k: axis_map[v] for k, v in new_axis_locations.items() if axis_map[v] is not None
    }
    new_dims = original_dims - len(removed_axis_indices)
    new_axis_locations = {
        k: (v - new_dims) if axis_negative[k] else v for k, v in new_axis_locations.items()
    }
    return new_axis_locations



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from collections import defaultdict
from itertools import chain

from compiler_config.config import InlineResultsProcessing

from qat.core.pass_base import LoweringPass
from qat.ir.instruction_builder import InstructionBuilder
from qat.ir.instructions import (
    Assign,
    GreaterThan,
    Instruction,
    Jump,
    Label,
    LoopCount,
    QuantumInstruction,
    Repeat,
    ResultsProcessing,
    Return,
    Synchronize,
    Variable,
)
from qat.ir.lowered import PartitionedIR
from qat.ir.measure import Acquire, PostProcessing


class PartitionByPulseChannel(LoweringPass):
    """Separates a list of instructions into their respective pulse channels.

    For targets that do not have native support for
    :class:`Synchronize <qat.purr.compiler.instructions.Synchronize>`, we can separate the
    instruction list into many lists of instructions, with each list only containing
    instructions that act on a single channel.

    This pass achieves the same as the :class:`TriagePass <qat.backend.passes.legacy.analysis.TriagePass>`,
    but instead of saving the results in the results manager, returns the partitioned
    instructions. The only reason I've changed the name is to not confuse it with the
    TriagePass :)
    """

    def run(self, ir: InstructionBuilder, *args, **kwargs) -> PartitionedIR:
        """
        :param ir: The list of instructions stored in a :class:`InstructionBuilder`.
        :param res_mgr: The result manager to save the analysis results.
        """
        shared_instructions = list()
        variables = defaultdict(list)
        partitioned_ir = PartitionedIR()

        if hasattr(ir, "shots") and hasattr(ir, "compiled_shots"):
            partitioned_ir.shots = ir.shots
            partitioned_ir.compiled_shots = ir.compiled_shots

        def add_to_all(inst: Instruction) -> None:
            shared_instructions.append(inst)
            for target_list in partitioned_ir.target_map.values():
                target_list.append(inst)
            partitioned_ir.target_map.default_factory = shared_instructions.copy

        partitioned_ir.repetition_period = getattr(ir, "repetition_period", None)

        partitioned_ir.passive_reset_time = getattr(ir, "passive_reset_time", None)

        repeat_seen = False
        for instr in ir:
            handled = False
            if isinstance(instr, QuantumInstruction):
                if isinstance(instr, Synchronize):
                    raise ValueError(
                        f"The `Synchronize` instruction {instr} is not supported by the "
                        "PartitionByPulseChannel pass. Please lower it to `Delay` instructions "
                        "using the `LowerSyncsToDelays` pass"
                    )
                handled = True
                for target in instr.targets:
                    partitioned_ir.target_map[target].append(instr)
            if isinstance(instr, Return):
                partitioned_ir.returns.append(instr)

            elif isinstance(instr, Variable):
                variables[instr.name].append(instr)
                add_to_all(instr)

            elif isinstance(instr, Assign):
                var_refs = variables[instr.name]
                if len(var_refs) == 0:
                    variables[instr.name].append(Variable(name=instr.name))
                if var_refs[-1].var_type == LoopCount:
                    add_to_all(instr)
                if isinstance(instr.value, list | str):
                    # Only list assigns that are readout/result assignements
                    # TODO: Supper different assigns for result readout vs loop
                    # variables etc.
                    partitioned_ir.assigns.append(instr)

            elif isinstance(instr, Acquire):
                for target in instr.targets:
                    partitioned_ir.acquire_map[target].append(instr)

            elif isinstance(instr, PostProcessing):
                partitioned_ir.pp_map[instr.output_variable].append(instr)

            elif isinstance(instr, ResultsProcessing):
                partitioned_ir.rp_map[instr.variable] = instr

            elif isinstance(instr, Repeat):
                if repeat_seen:
                    raise ValueError("Multiple Repeat instructions found.")
                if partitioned_ir.shots is None:
                    partitioned_ir.shots = instr.repeat_count
                    partitioned_ir.compiled_shots = instr.repeat_count
                repeat_seen = True

            elif isinstance(instr, Jump):
                if isinstance(instr.condition, GreaterThan):
                    if repeat_seen:
                        raise ValueError("Multiple `Repeat` or `Jump` instructions found.")
                    if partitioned_ir.shots is None:
                        limit = instr.condition.left
                        partitioned_ir.shots = limit
                        partitioned_ir.compiled_shots = limit
                add_to_all(instr)

            elif isinstance(instr, Label):
                add_to_all(instr)
            elif not handled:
                raise TypeError(f"Unexpected Instruction type: {type(instr)}.")

        for pulse_channel_id in partitioned_ir.target_map:
            partitioned_ir.pulse_channels[pulse_channel_id] = ir.get_pulse_channel(
                pulse_channel_id
            )

        # Assume that raw acquisitions are experiment results.
        # TODO: separate as ResultsProcessingSanitisation pass. (COMPILER-412)
        acquires = list(chain(*partitioned_ir.acquire_map.values()))
        missing_results = {
            acq.output_variable
            for acq in acquires
            if acq.output_variable not in partitioned_ir.rp_map
        }
        for missing_var in missing_results:
            partitioned_ir.rp_map[missing_var] = ResultsProcessing(
                variable=missing_var, results_processing=InlineResultsProcessing.Experiment
            )

        return partitioned_ir


PydPartitionByPulseChannel = PartitionByPulseChannel



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from qat.pipelines.pipeline import Pipeline
from qat.pipelines.updateable import UpdateablePipeline


class RTCSPipeline(UpdateablePipeline):
    """A pipeline that compiles programs using the real time chip simulator."""

    @staticmethod
    def _build_pipeline(config, model, target_data=None, engine=None) -> Pipeline:
        raise NotImplementedError("This pipeline is not implemented yet")



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd
from __future__ import annotations

import re
from typing import Any

import numpy as np
from pydantic import Field, field_validator, model_validator

from qat.ir.waveforms import GaussianWaveform, SoftSquareWaveform
from qat.utils.pydantic import (
    CalibratablePositiveFloat,
    ComplexNDArray,
    FloatNDArray,
    FrozenDict,
    NoExtraFieldsModel,
    QubitId,
    WaveformType,
)
from qat.utils.uuid import uuid4


class Component(NoExtraFieldsModel):
    """
    Base class for any logical object which can act as a target of a quantum action
    - a Qubit or various channels for a simple example.

    :param uuid: The unique string representation of the component.
    """

    uuid: str = Field(default_factory=lambda: str(uuid4()), frozen=True)

    def __hash__(self):
        return hash(self.uuid)

    def __eq__(self, other: Component):
        if self.uuid != other.uuid:
            return False

        if type(self) is not type(other):
            return False

        if self.__class__.model_fields != other.__class__.model_fields:
            return False

        for field_name in self.__class__.model_fields:
            field_s = getattr(self, field_name)
            field_o = getattr(other, field_name)
            if isinstance(field_s, float) and isinstance(field_o, float):
                if np.isnan(field_s) and np.isnan(field_o):
                    continue

            if field_s != field_o:
                return False

        return True

    def __repr__(self):
        return f"{self.__class__.__name__}({self.uuid})"

    def __str__(self):
        return self.__repr__()

    @property
    def is_calibrated(self):
        for field_name in self.__class__.model_fields:
            field_value = getattr(self, field_name)
            if (
                isinstance(field_value, (Component, PulseChannelSet))
                and not field_value.is_calibrated
            ):
                return False
            elif isinstance(field_value, float) and np.isnan(field_value):
                return False
        return True


class PhysicalBaseband(Component):
    """
    Models the Local Oscillator (LO) used with a mixer to change the
    frequency of a carrier signal.

    :param frequency: The frequency of the LO.
    :param if_frequency: The intermediate frequency (IF) resulting from
                      mixing the baseband with the carrier signal.
    """

    frequency: CalibratablePositiveFloat = Field(default=np.nan)
    if_frequency: CalibratablePositiveFloat = Field(default=np.nan)


class IQBias(NoExtraFieldsModel):
    """
    Models the bias in voltages V_I / V_Q for the I and Q components in a physical channel.

    :param bias: The bias as a complex number, where the real part is the I component
                and the imaginary part is the Q component.
    """

    bias: complex | float = 0 + 0j


class PhysicalChannel(Component):
    """
    Models a physical channel that can carry one or multiple pulses.

    Attributes:
        baseband: The physical baseband.

        block_size: The number of samples within a single block.
        iq_voltage_bias: The bias in voltages V_I / V_Q for the I and Q components.

        acquire_allowed: Whether the physical channel allows acquire operations.
        swap_readout_iq: Whether to swap the I and Q components for readout operations.
        name_index: integer used in human readable name, `channel_<index>`.
    """

    baseband: PhysicalBaseband = Field(frozen=True)

    block_size: int | None = Field(ge=1, default=1)
    iq_voltage_bias: IQBias = Field(default=IQBias())
    name_index: int = Field(ge=0, frozen=True)

    @property
    def I_bias(self) -> float:
        return self.iq_voltage_bias.bias.real

    @property
    def Q_bias(self) -> float:
        return self.iq_voltage_bias.bias.imag


class QubitPhysicalChannel(PhysicalChannel): ...


class ResonatorPhysicalChannel(PhysicalChannel):
    swap_readout_iq: bool = False


pulse_channel_check = re.compile(r"PulseChannel$")


class PulseChannel(Component):
    """
    Models a pulse channel on a particular device.

    :param physical_channel: Physical channel that carries the pulse.
    :param frequency: Frequency of the pulse.
    :param imbalance: Ratio between I and Q AC voltage that is sent out of the FPGA.
    :param phase_iq_offset: Phase offset between the I and Q plane for AC voltage that is send out the FPGA.
    :param scale: Scale factor for mapping the voltage of the pulse to frequencies.
    """

    frequency: CalibratablePositiveFloat = Field(default=np.nan)
    imbalance: float = 1.0
    phase_iq_offset: float = 0.0
    scale: complex | float = 1.0 + 0.0j

    def model_post_init(self, context: Any, /) -> None:
        if isinstance(self.scale, np.complexfloating):
            self.scale = complex(self.scale)

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if not pulse_channel_check.search(cls.__name__):
            raise TypeError(
                f"The class name '{cls.__name__}' must contain `PulseChannel` at the end."
            )

    @property
    def pulse_type(self):
        return self.__class__.__name__.replace("PulseChannel", "").lower()


class QubitPulseChannel(PulseChannel): ...


class ResonatorPulseChannel(PulseChannel): ...


class CalibratablePulse(NoExtraFieldsModel):
    waveform_type: WaveformType = GaussianWaveform
    width: CalibratablePositiveFloat = Field(default=100e-09, ge=0)
    amp: float = 0.25 / (100e-9 * 1.0 / 3.0 * np.pi**0.5)
    phase: float = 0.0
    drag: float = 0.0
    rise: float = 1.0 / 3.0
    amp_setup: float = 0.0
    std_dev: float = 0.0


class CalibratableAcquire(NoExtraFieldsModel):
    delay: CalibratablePositiveFloat = Field(default=180e-08, ge=0)
    width: CalibratablePositiveFloat = Field(default=1e-06, ge=0)
    sync: bool = True
    weights: ComplexNDArray | FloatNDArray = Field(
        default_factory=lambda: ComplexNDArray([])
    )
    use_weights: bool = False


class DrivePulseChannel(QubitPulseChannel):
    """
    The pulse channel that drives the qubit from |0> -> |1>.

    :param pulse: Calibratable parameters for the X(pi/2) drive pulse.
    """

    pulse: CalibratablePulse = Field(
        default=CalibratablePulse(
            waveform_type=GaussianWaveform, width=100e-9, rise=1.0 / 3.0
        ),
        frozen=True,
    )

    pulse_x_pi: CalibratablePulse | None = Field(default=None)


class MeasurePulseChannel(ResonatorPulseChannel):
    """
    The pulse channel that measures the quantum state of the resonator.

    :param pulse: Calibratable parameters for the measure pulse.
    """

    pulse: CalibratablePulse = Field(default=CalibratablePulse(width=1e-06), frozen=True)


class AcquirePulseChannel(ResonatorPulseChannel):
    acquire: CalibratableAcquire = Field(default=CalibratableAcquire(), frozen=True)


class MeasureAcquirePulseChannel(MeasurePulseChannel, AcquirePulseChannel): ...


class SecondStatePulseChannel(QubitPulseChannel):
    active: bool = False
    delay: float = 0.0

    pulse: CalibratablePulse = Field(
        default=CalibratablePulse(
            waveform_type=GaussianWaveform, width=100e-9, rise=1.0 / 3.0
        ),
        frozen=True,
    )


class FreqShiftPulseChannel(QubitPulseChannel):
    active: bool = False
    amp: float = 1.0
    phase: float = 0.0


class CrossResonancePulseChannel(QubitPulseChannel):
    auxiliary_qubit: QubitId
    zx_pi_4_pulse: CalibratablePulse = Field(
        default=CalibratablePulse(
            waveform_type=SoftSquareWaveform, width=125e-9, rise=10e-9, amp=1e6
        ),
        frozen=True,
    )

    def __repr__(self):
        return f"{self.__class__.__name__}(@Q{self.auxiliary_qubit})"


class CrossResonanceCancellationPulseChannel(QubitPulseChannel):
    auxiliary_qubit: QubitId

    def __repr__(self):
        return f"{self.__class__.__name__}(@Q{self.auxiliary_qubit})"


class PulseChannelSet(NoExtraFieldsModel):
    """
    Encapsulates a set of pulse channels.
    """

    @property
    def is_calibrated(self):
        for field_name in self.__class__.model_fields:
            field_value = getattr(self, field_name)
            if isinstance(field_value, Component) and not field_value.is_calibrated:
                return False
            elif isinstance(field_value, tuple):
                for pulse_channel in field_value:
                    if (
                        isinstance(pulse_channel, Component)
                        and not pulse_channel.is_calibrated
                    ):
                        return False
        return True

    def __eq__(self, other: PulseChannelSet):
        if type(self) is not type(other):
            return False

        if self.__class__.model_fields != other.__class__.model_fields:
            return False

        for field_name in self.__class__.model_fields:
            if getattr(self, field_name) != getattr(other, field_name):
                return False

        return True

    def __ne__(self, other: PulseChannelSet):
        return not self.__eq__(other)

    def pulse_channel_with_id(self, id_: str):
        for field_name in self.__class__.model_fields:
            if (
                isinstance(comp := getattr(self, field_name), Component)
                and comp.uuid == id_
            ):
                return comp

        return None


class ResonatorPulseChannels(PulseChannelSet):
    measure: MeasurePulseChannel = Field(default=MeasurePulseChannel())
    acquire: AcquirePulseChannel = Field(default=AcquirePulseChannel())

    @property
    def all_pulse_channels(self):
        return [self.measure, self.acquire]


class Resonator(Component):
    """
    Models a resonator on a chip. Can be connected to multiple qubits.

    :param physical_channel: The physical channel that carries the pulses to the physical resonator.
    :param pulse_channels: The pulse channels for controlling the resonator.
    :param measure_pulse: Calibrated parameters for the measure pulse on the resonator.
    :param acquire: Calibrated parameters for the acquire instruction.
    """

    physical_channel: ResonatorPhysicalChannel
    pulse_channels: ResonatorPulseChannels = Field(
        frozen=True, default=ResonatorPulseChannels()
    )

    @property
    def all_pulse_channels(self):
        return self.pulse_channels.all_pulse_channels

    @property
    def measure_pulse_channel(self):
        return self.pulse_channels.measure

    @property
    def acquire_pulse_channel(self):
        return self.pulse_channels.acquire


class QubitPulseChannels(PulseChannelSet):
    drive: DrivePulseChannel = Field(frozen=True, default=DrivePulseChannel())
    second_state: SecondStatePulseChannel = Field(
        frozen=True, default=SecondStatePulseChannel()
    )
    freq_shift: FreqShiftPulseChannel = Field(frozen=True, default=FreqShiftPulseChannel())

    cross_resonance_channels: FrozenDict[QubitId, CrossResonancePulseChannel] = Field(
        frozen=True, max_length=3, default=FrozenDict({})
    )
    cross_resonance_cancellation_channels: FrozenDict[
        QubitId, CrossResonanceCancellationPulseChannel
    ] = Field(frozen=True, max_length=3, default=FrozenDict({}))

    @field_validator("cross_resonance_channels", "cross_resonance_cancellation_channels")
    def validate_channels_qubit_mapping(cls, channels):
        for aux_qubit_id, pulse_channel in channels.items():
            assert aux_qubit_id == pulse_channel.auxiliary_qubit, (
                f"Mismatch in mapping for qubit id in {channels}."
            )
        return channels

    @model_validator(mode="after")
    def validate_cross_resonance_pulse_channels(self):
        assert (
            self.cross_resonance_channels.keys()
            == self.cross_resonance_cancellation_channels.keys()
        ), (
            "Mismatch between auxiliary qubit ids for cross resonance and cross resonance cancellation channels."
        )
        return self

    @property
    def all_pulse_channels(self):
        return [
            self.drive,
            self.second_state,
            self.freq_shift,
            *self.cross_resonance_channels.values(),
            *self.cross_resonance_cancellation_channels.values(),
        ]

    def pulse_channel_with_id(self, id_: str):
        for pulse_ch in self.all_pulse_channels:
            if pulse_ch.uuid == id_:
                return pulse_ch

        return None


class Qubit(Component):
    """
    Models a superconducting qubit on a chip, and holds all information relating to it.

    :param physical_channel: The physical channel that carries the pulses to the physical qubit.
    :param pulse_channels: The pulse channels for controlling the qubit.
    :param resonator: The measure device of the qubit.
    :param pulse: Calibrated parameters for the X(pi/2) pulse.
    """

    physical_channel: QubitPhysicalChannel
    pulse_channels: QubitPulseChannels = Field(frozen=True, default=QubitPulseChannels())
    resonator: Resonator

    mean_z_map_args: list[complex | float] = Field(max_length=2, default=[1.0, 0.0])
    discriminator: complex | float = 0.0

    direct_x_pi: bool = False

    def model_post_init(self, context: Any, /) -> None:
        if isinstance(self.discriminator, np.complexfloating):
            self.discriminator = complex(self.discriminator)

        for i in range(len(self.mean_z_map_args)):
            if isinstance(self.mean_z_map_args[i], np.complexfloating):
                self.mean_z_map_args[i] = complex(self.mean_z_map_args[i])

    @property
    def all_pulse_channels(self):
        return self.pulse_channels.all_pulse_channels

    @property
    def all_qubit_and_resonator_pulse_channels(self):
        return self.all_pulse_channels + self.resonator.all_pulse_channels

    @property
    def drive_pulse_channel(self):
        return self.pulse_channels.drive

    @property
    def second_state_pulse_channel(self):
        return self.pulse_channels.second_state

    @property
    def freq_shift_pulse_channel(self):
        return self.pulse_channels.freq_shift

    @property
    def cross_resonance_pulse_channels(self):
        return self.pulse_channels.cross_resonance_channels

    @property
    def cross_resonance_cancellation_pulse_channels(self):
        return self.pulse_channels.cross_resonance_cancellation_channels

    @property
    def measure_pulse_channel(self):
        return self.resonator.pulse_channels.measure

    @property
    def acquire_pulse_channel(self):
        return self.resonator.pulse_channels.acquire



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
import re
from abc import ABC, abstractmethod
from pathlib import Path

from compiler_config.config import CompilerConfig, InlineResultsProcessing

from qat.core.metrics_base import MetricsManager
from qat.core.pass_base import PassManager
from qat.core.result_base import ResultManager
from qat.frontend.base import BaseFrontend
from qat.frontend.parsers.qasm import (
    CloudQasmParser as PydCloudQasmParser,
    Qasm3Parser as PydQasm3Parser,
)
from qat.frontend.parsers.qasm.base import AbstractParser as PydAbstractParser
from qat.frontend.passes.analysis import InputAnalysis
from qat.frontend.passes.purr.transform import InputOptimisation
from qat.frontend.passes.transform import PydInputOptimisation
from qat.ir.builder_factory import BuilderFactory
from qat.model.hardware_model import PhysicalHardwareModel as PydHardwareModel
from qat.purr.compiler.builders import InstructionBuilder
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.purr.integrations.qasm import AbstractParser, CloudQasmParser, Qasm3Parser

path_regex = re.compile(r"^.+\.(qasm)$")
string_regex = re.compile(r"OPENQASM (?P<version>[0-9]+)(?:.[0-9])?;")


def is_qasm_path(src: str) -> bool:
    """Determines if the string corresponds to a path, and has a `.qasm` extension.

    :param src: The path to the source program.
    """

    if path_regex.match(src) is not None:
        return True
    return False


def is_qasm_str(src: str) -> bool:
    """Determines if the string contains OPENQASM headers.

    :param src: The source program as a string.
    """

    if string_regex.search(src) is not None:
        return True
    return False


def get_qasm_version(src: str) -> int:
    """Extracts the QASM version from the source file.

    :param: The source program as a string.
    :returns: The major version number.
    """

    return int(string_regex.search(src).group("version"))


def load_qasm_file(path: str) -> str:
    """Loads a Qasm file from a given file path.

    :param path: The file path.
    :raises ValueError: Raises an error if the the file has an invalid extension.
    :returns: The Qasm program as a string or bytes.
    """

    path = Path(path)
    if path.suffix == ".qasm":
        with path.open() as file:
            src = file.read()
    else:
        raise ValueError("String expected to end in `.qasm`.")
    return src


class BaseQasmFrontend(BaseFrontend, ABC):
    """A base frontend for QASM programs.

    Handles the parsing of QASM into QAT's intermediate representation (IR). Optionally, it
    can also run pipelines before to optimize and validate QASM files. The QASM2 and QASM3
    frontends are identical up to the language and its respective parser. This class
    implements the base functionality."""

    def __init__(
        self,
        model: QuantumHardwareModel | PydHardwareModel = None,
        pipeline: None | PassManager = None,
    ):
        """
        :param model: The hardware model can be required for the pipeline and is used in
            parsing.
        :param pipeline: A pipeline for validation and optimising QASM, defaults to a
            predefined pipeline that optimizes the QASM file.
        """
        super().__init__(model)
        self.pipeline = (
            pipeline if pipeline is not None else self.build_pass_pipeline(model)
        )

    @property
    @abstractmethod
    def version(self) -> int: ...

    @abstractmethod
    def create_parser(
        self, config: CompilerConfig
    ) -> AbstractParser | PydAbstractParser: ...

    @staticmethod
    def build_pass_pipeline(model: QuantumHardwareModel) -> PassManager | None:
        """Creates a pipeline to optimize QASM files.

        :param model: The hardware model is needed for optimizations.
        :return: The pipeline as a pass manager.
        """

        # TODO: replace input analysis + input optimisation with a qasm specific pass?
        # (COMPILER-340)
        if isinstance(model, QuantumHardwareModel):  # legacy hardware model
            return PassManager() | InputAnalysis() | InputOptimisation(model)
        elif isinstance(model, PydHardwareModel):  # pydantic hardware model
            return PassManager() | InputAnalysis() | PydInputOptimisation(model)

    def check_and_return_source(self, src: str) -> bool | str:
        """Checks that the source program (or file path) can be interpreted as a QASM file
        by checking against the language.

        :param src: The QASM program, or the file path to the program.
        :returns: If the program is determined to not be valid, False is returned.
            Otherwise, the program is returned (and loaded if required).
        """

        if not isinstance(src, str):
            return False

        if is_qasm_path(src):
            if Path(src).is_file():
                src = load_qasm_file(src)

        if not is_qasm_str(src):
            return False

        if get_qasm_version(src) != self.version:
            return False

        return src

    def _check_source(self, src: str) -> str:
        """
        Checks that the source program (or file path) can be interpreted as a QASM file and raises and error if not.
        """
        src = self.check_and_return_source(src)
        if not src:
            raise ValueError(
                f"Source program is not a QASM{self.version} program, or a path "
                f"to a QASM{self.version} program."
            )
        return src

    def _generate_builder(
        self,
        src: str,
        compiler_config: CompilerConfig,
    ) -> InstructionBuilder | None:
        """
        Creates the instruction builder for execution from the base builder and compiler config.
        :param builder: The base instruction builder.
        :param compiler_config: The compiler config is used in both the pipeline and for
            parsing.
        """
        builder = BuilderFactory.create_builder(self.model)
        if isinstance(self.model, PydHardwareModel):
            builder.repeat(compiler_config.repeats)
        elif isinstance(self.model, QuantumHardwareModel):
            builder = builder.repeat(
                compiler_config.repeats,
                repetition_period=compiler_config.repetition_period,
                passive_reset_time=compiler_config.passive_reset_time,
            )

        parser = self.create_parser(compiler_config)
        builder = parser.parse(builder, src)
        return builder

    def emit(
        self,
        src: str,
        res_mgr: ResultManager | None = None,
        met_mgr: MetricsManager | None = None,
        compiler_config: CompilerConfig | None = None,
        **kwargs,
    ) -> InstructionBuilder:
        """Compiles the source QASM program into QAT's intermediate representation.

        :param src: The source program, or path to the source program.
        :param res_mgr: The result manager to store results from passes within the pipeline,
            defaults to an empty :class:`ResultManager`.
        :param met_mgr: The metrics manager to store metrics, such as the optimized QASM
            circuit, defaults to an empty :class:`MetricsManager`.
        :param compiler_config: The compiler config is used in both the pipeline and for
            parsing.
        :raises ValueError: An error is thrown if the the program is not detected to be a
            QASM program.
        :returns: The program as an :class:`InstructionBuilder`.
        """
        res_mgr, met_mgr, compiler_config = self._check_metrics_and_config(
            res_mgr, met_mgr, compiler_config
        )

        src = self._check_source(src)
        src = self.pipeline.run(
            src, res_mgr, met_mgr, compiler_config=compiler_config, **kwargs
        )
        return self._generate_builder(src, compiler_config)


class Qasm2Frontend(BaseQasmFrontend):
    """A frontend for QASM2 programs.

    Handles the parsing of QASM2 into QATs intermediate representation. Optionally, it can
    also run pipelines before to optimize and validate QASM2 files.
    """

    version = 2

    def __init__(self, model: QuantumHardwareModel, pipeline: None | PassManager = None):
        """
        :param model: The hardware model can be required for the pipeline and is used in
            parsing.
        :param pipeline: A pipeline for validation and optimising QASM, defaults to a
            predefined pipeline that optimizes the QASM file.
        """

        super().__init__(model, pipeline)

    def create_parser(self, config: CompilerConfig) -> CloudQasmParser | PydCloudQasmParser:
        if isinstance(self.model, QuantumHardwareModel):
            parser = CloudQasmParser()
        elif isinstance(self.model, PydHardwareModel):
            parser = PydCloudQasmParser()
        else:
            raise TypeError("Invalid hardware model type set.")

        parser.results_format = (
            config.results_format.format
            if config.results_format.format is not None
            else InlineResultsProcessing.Program
        )
        return parser


class Qasm3Frontend(BaseQasmFrontend):
    """A frontend for QASM3 programs.

    Handles the parsing of QASM3 into QATs intermediate representation. Optionally, it can
    also run pipelines before to optimize and validate QASM3 files.
    """

    version = 3

    def __init__(self, model: QuantumHardwareModel, pipeline: None | PassManager = None):
        """
        :param model: The hardware model can be required for the pipeline and is used in
            parsing.
        :param pipeline: A pipeline for validation and optimising QASM, defaults to a
            predefined pipeline that optimizes the QASM file.
        """

        super().__init__(model, pipeline)

    def create_parser(self, config: CompilerConfig) -> Qasm3Parser | PydQasm3Parser:
        if isinstance(self.model, QuantumHardwareModel):
            parser = Qasm3Parser()
        elif isinstance(self.model, PydHardwareModel):
            parser = PydQasm3Parser()
        else:
            raise TypeError("Invalid hardware model type set.")

        parser.results_format = (
            config.results_format.format
            if config.results_format.format is not None
            else InlineResultsProcessing.Program
        )
        return parser



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from compiler_config.config import CompilerConfig

from qat.backend.base import BaseBackend
from qat.core.metrics_base import MetricsManager
from qat.core.result_base import ResultManager
from qat.purr.compiler.builders import InstructionBuilder


class FallthroughBackend(BaseBackend):
    """
    A backend that passes through an IR :class:`InstructionBuilder` and does not alter it.
    """

    def __init__(self, model: None = None):
        """
        :param model: The hardware model that holds calibrated information on the qubits on
            the QPU.
        """
        self.model = model

    def emit(
        self,
        ir: InstructionBuilder,
        res_mgr: ResultManager | None = None,
        met_mgr: MetricsManager | None = None,
        compiler_config: CompilerConfig | None = None,
        **kwargs,
    ):
        return ir



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Dict, List, Optional

from qblox_instruments.qcodes_drivers.module import Module
from qblox_instruments.qcodes_drivers.sequencer import Sequencer

from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


@dataclass
class ConnectionConfig:
    bulk_value: List[str] = field(default_factory=list)
    out0: Optional[str] = None  # Enum: {‘I’, ‘Q’, ‘off’}
    out1: Optional[str] = None  # Enum: {‘I’, ‘Q’, ‘off’}
    out2: Optional[str] = None  # Enum: {‘I’, ‘Q’, ‘off’}
    out3: Optional[str] = None  # Enum: {‘I’, ‘Q’, ‘off’}
    acq: Optional[str] = None  # Enum: {False, ‘in0’, True, ‘off’}
    acq_I: Optional[str] = None  # Enum: {‘in0’, ‘in1’, ‘off’}
    acq_Q: Optional[str] = None  # Enum: {‘in0’, ‘in1’, ‘off’}


@dataclass
class NcoConfig:
    freq: Optional[float] = None
    phase_offs: Optional[float] = None
    prop_delay_comp: Optional[int] = None
    prop_delay_comp_en: Optional[bool] = None


@dataclass
class AwgConfig:
    cont_mode_en_path0: Optional[bool] = None
    cont_mode_en_path1: Optional[bool] = None
    cont_mode_waveform_idx_path0: Optional[int] = None
    cont_mode_waveform_idx_path1: Optional[int] = None

    upsample_rate_path0: Optional[int] = None
    upsample_rate_path1: Optional[int] = None

    gain_path0: Optional[float] = None
    gain_path1: Optional[float] = None

    offset_path0: Optional[float] = None
    offset_path1: Optional[float] = None

    mod_en: Optional[bool] = None


@dataclass
class MixerConfig:
    phase_offset: Optional[float] = None
    gain_ratio: Optional[float] = None


@dataclass
class SquareWeightAcq:
    integration_length: Optional[int] = None


@dataclass
class ThresholdedAcqConfig:
    rotation: Optional[float] = None
    threshold: Optional[float] = None
    marker_en: Optional[bool] = None
    marker_address: Optional[int] = None
    marker_invert: Optional[bool] = None
    trigger_en: Optional[bool] = None
    trigger_address: Optional[int] = None
    trigger_invert: Optional[bool] = None


@dataclass
class TtlAcqConfig:
    auto_bin_incr_en: Optional[bool] = None
    threshold: Optional[float] = None
    input_select: Optional[int] = None


@dataclass
class SequencerConfig:
    sync_en: Optional[bool] = None
    marker_ovr_en: Optional[bool] = None
    marker_ovr_value: Optional[int] = None

    trigger_count_thresholds: Dict[int, float] = field(
        default_factory=dict
    )  # key in [0,15]
    trigger_threshold_inverts: Dict[int, bool] = field(
        default_factory=dict
    )  # key in [0,15]

    connection: ConnectionConfig = field(default_factory=lambda: ConnectionConfig())
    nco: NcoConfig = field(default_factory=lambda: NcoConfig())
    awg: AwgConfig = field(default_factory=lambda: AwgConfig())
    mixer: MixerConfig = field(default_factory=lambda: MixerConfig())

    demod_en_acq: Optional[bool] = None
    square_weight_acq: SquareWeightAcq = field(default_factory=lambda: SquareWeightAcq())
    thresholded_acq: ThresholdedAcqConfig = field(
        default_factory=lambda: ThresholdedAcqConfig()
    )
    ttl_acq: TtlAcqConfig = field(default_factory=lambda: TtlAcqConfig())


@dataclass
class OffsetConfig:
    out0: Optional[float] = None
    out1: Optional[float] = None
    out2: Optional[float] = None
    out3: Optional[float] = None

    in1: Optional[float] = None
    in0: Optional[float] = None

    out0_path0: Optional[float] = None
    out0_path1: Optional[float] = None
    out1_path0: Optional[float] = None
    out1_path1: Optional[float] = None

    in0_path0: Optional[float] = None
    in0_path1: Optional[float] = None


@dataclass
class LoConfig:
    out0_en: Optional[bool] = None
    out0_freq: Optional[float] = None
    out1_en: Optional[bool] = None
    out1_freq: Optional[float] = None

    out0_in0_en: Optional[bool] = None
    out0_in0_freq: Optional[float] = None


@dataclass
class AttConfig:
    out0: Optional[int] = None
    out1: Optional[int] = None
    in0: Optional[int] = None


@dataclass
class GainConfig:
    in0: Optional[int] = None
    in1: Optional[int] = None


@dataclass
class ScopeAcqConfig:
    sequencer_select: Optional[int] = None
    trigger_mode_path0: Optional[str] = None  # Enum: {‘sequencer’, ‘level’}
    trigger_mode_path1: Optional[str] = None  # Enum: {‘sequencer’, ‘level’}
    trigger_level_path0: Optional[float] = None
    trigger_level_path1: Optional[float] = None
    avg_mode_en_path0: Optional[bool] = None
    avg_mode_en_path1: Optional[bool] = None


@dataclass
class ModuleConfig:
    marker_inverts: Dict[int, bool] = field(default_factory=dict)

    offset: OffsetConfig = field(default_factory=lambda: OffsetConfig())
    lo: LoConfig = field(default_factory=lambda: LoConfig())
    attenuation: AttConfig = field(default_factory=lambda: AttConfig())
    gain: GainConfig = field(default_factory=lambda: GainConfig())
    scope_acq: ScopeAcqConfig = field(default_factory=lambda: ScopeAcqConfig())


@dataclass
class QbloxConfig:
    slot_idx: Optional[int] = None
    module: ModuleConfig = field(default_factory=lambda: ModuleConfig())
    sequencers: Dict[int, SequencerConfig] = field(default_factory=dict)


class QbloxConfigHelper(ABC):
    def __init__(
        self,
        module_config: ModuleConfig = None,
        sequencer_config: SequencerConfig = None,
    ):
        self.sequencer_config = sequencer_config or SequencerConfig()
        self.module_config = module_config or SequencerConfig()

    def configure(self, module: Module, sequencer: Sequencer):
        self.configure_module(module)
        self.configure_sequencer(sequencer)

    @abstractmethod
    def configure_module(self, module: Module):
        pass

    @abstractmethod
    def configure_sequencer(self, sequencer: Sequencer):
        pass

    def configure_connection(self, sequencer: Sequencer):
        if self.sequencer_config.connection.bulk_value:
            sequencer.connect_sequencer(*self.sequencer_config.connection.bulk_value)

    def configure_nco(self, sequencer: Sequencer):
        if self.sequencer_config.nco.freq:
            sequencer.nco_freq(self.sequencer_config.nco.freq)
        if self.sequencer_config.nco.prop_delay_comp_en:
            sequencer.nco_prop_delay_comp_en(self.sequencer_config.nco.prop_delay_comp_en)
        if self.sequencer_config.nco.prop_delay_comp:
            sequencer.nco_prop_delay_comp(self.sequencer_config.nco.prop_delay_comp)
        if self.sequencer_config.nco.phase_offs:
            sequencer.nco_phase_offs(self.sequencer_config.nco.phase_offs)

    def configure_awg(self, sequencer: Sequencer):
        if self.sequencer_config.awg.mod_en:
            sequencer.mod_en_awg(self.sequencer_config.awg.mod_en)
        if self.sequencer_config.awg.gain_path0:
            sequencer.gain_awg_path0(self.sequencer_config.awg.gain_path0)
        if self.sequencer_config.awg.gain_path1:
            sequencer.gain_awg_path1(self.sequencer_config.awg.gain_path1)
        if self.sequencer_config.awg.offset_path0:
            sequencer.offset_awg_path0(self.sequencer_config.awg.offset_path0)
        if self.sequencer_config.awg.offset_path1:
            sequencer.offset_awg_path1(self.sequencer_config.awg.offset_path1)

        if self.sequencer_config.awg.cont_mode_en_path0:
            sequencer.cont_mode_en_awg_path0(self.sequencer_config.awg.cont_mode_en_path0)
        if self.sequencer_config.awg.cont_mode_en_path1:
            sequencer.cont_mode_en_awg_path1(self.sequencer_config.awg.cont_mode_en_path1)
        if self.sequencer_config.awg.cont_mode_waveform_idx_path0:
            sequencer.cont_mode_waveform_idx_awg_path0(
                self.sequencer_config.awg.cont_mode_waveform_idx_path0
            )
        if self.sequencer_config.awg.cont_mode_waveform_idx_path1:
            sequencer.cont_mode_waveform_idx_awg_path1(
                self.sequencer_config.awg.cont_mode_waveform_idx_path1
            )

        if self.sequencer_config.awg.cont_mode_en_path0:
            sequencer.cont_mode_en_awg_path0(self.sequencer_config.awg.cont_mode_en_path0)
        if self.sequencer_config.awg.cont_mode_en_path1:
            sequencer.cont_mode_en_awg_path1(self.sequencer_config.awg.cont_mode_en_path1)
        if self.sequencer_config.awg.cont_mode_waveform_idx_path0:
            sequencer.cont_mode_waveform_idx_awg_path0(
                self.sequencer_config.awg.cont_mode_waveform_idx_path0
            )
        if self.sequencer_config.awg.cont_mode_waveform_idx_path1:
            sequencer.cont_mode_waveform_idx_awg_path1(
                self.sequencer_config.awg.cont_mode_waveform_idx_path1
            )

    def configure_mixer(self, sequencer: Sequencer):
        if self.sequencer_config.mixer.phase_offset:
            sequencer.mixer_corr_phase_offset_degree(
                self.sequencer_config.mixer.phase_offset
            )
        if self.sequencer_config.mixer.gain_ratio:
            sequencer.mixer_corr_gain_ratio(self.sequencer_config.mixer.gain_ratio)

    def calibrate_mixer(
        self, module: Module, sequencer: Sequencer = None, connection: str = None
    ):
        """
        - Compensates for the LO leakage for input(s)/output(s) depending on module
        - Suppresses undesired sideband on sequencer

        :param module: A Module instance representing a QCM-RF or a QRM-RF card.
        :param sequencer: A Sequencer instance representing the sequencer.
            When None, sideband cal will run on all sequencers belonging to module `module`
        :param connection: An optional string indicating muxing of Sequencer to analog channel
            Defaults to "out0"
        """

        sequencers = [sequencer] if sequencer else module.sequencers

        if connection:
            log.info(
                f"Provided argument `connection` is {connection}. Will overwrite sequencer's connections"
            )
            module.disconnect_outputs()
            for sequencer in sequencers:
                sequencer.connect_sequencer(connection)
        else:
            log.info(
                "Argument `connection` not provided. Assuming `connection` already specified"
            )

        log.info(f"Calibrating LO leakage on module {module}")
        offset_config = self.calibrate_lo_leakage(module)
        mixer_configs = {}
        for sequencer in sequencers:
            mixer_configs[sequencer.seq_idx] = self.calibrate_sideband(
                sequencer, connection
            )

        if connection:
            log.info(
                f"Mixer calibration finished. Resetting sequencer connections for module {module.slot_idx}"
            )
            module.disconnect_outputs()

        return offset_config, mixer_configs

    @abstractmethod
    def calibrate_lo_leakage(self, module: Module):
        pass

    def calibrate_sideband(self, sequencer: Sequencer, connection: str = None):
        log.info(f"Calibrating sidebands on sequencer {sequencer}")
        sequencer.sideband_cal()
        sequencer.arm_sequencer()
        sequencer.start_sequencer()

        return MixerConfig(
            phase_offset=sequencer.mixer_corr_phase_offset_degree(),
            gain_ratio=sequencer.mixer_corr_gain_ratio(),
        )


class QcmConfigHelper(QbloxConfigHelper):
    pass


class QcmRfConfigHelper(QcmConfigHelper):
    def configure_module(self, module: Module):
        self.configure_lo(module)
        self.configure_attenuation(module)
        self.configure_offset(module)

    def configure_sequencer(self, sequencer: Sequencer):
        self.configure_connection(sequencer)
        self.configure_nco(sequencer)
        self.configure_awg(sequencer)
        self.configure_mixer(sequencer)

    def configure_lo(self, module: Module):
        if self.module_config.lo.out0_en:
            module.out0_lo_en(self.module_config.lo.out0_en)  # Switch the LO 0 on
        if self.module_config.lo.out0_freq:
            module.out0_lo_freq(self.module_config.lo.out0_freq)
        if self.module_config.lo.out1_en:
            module.out1_lo_en(self.module_config.lo.out1_en)  # Switch the LO 1 on
        if self.module_config.lo.out1_freq:
            module.out1_lo_freq(self.module_config.lo.out1_freq)

    def configure_attenuation(self, module: Module):
        if self.module_config.attenuation.out0:
            module.out0_att(self.module_config.attenuation.out0)
        if self.module_config.attenuation.out1:
            module.out1_att(self.module_config.attenuation.out1)

    def configure_offset(self, module: Module):
        if self.module_config.offset.out0_path0:
            module.out0_offset_path0(self.module_config.offset.out0_path0)
        if self.module_config.offset.out0_path1:
            module.out0_offset_path1(self.module_config.offset.out0_path1)
        if self.module_config.offset.out1_path0:
            module.out1_offset_path0(self.module_config.offset.out1_path0)
        if self.module_config.offset.out1_path1:
            module.out1_offset_path1(self.module_config.offset.out1_path1)

    def calibrate_lo_leakage(self, module: Module):
        module.out0_lo_cal()
        module.out1_lo_cal()

        return OffsetConfig(
            out0_path0=module.out0_offset_path0(),
            out0_path1=module.out0_offset_path1(),
            out1_path0=module.out1_offset_path0(),
            out1_path1=module.out1_offset_path1(),
        )


class QrmConfigHelper(QbloxConfigHelper):
    def configure_scope_acq(self, module: Module):
        scope_acq = self.module_config.scope_acq
        if scope_acq.sequencer_select:
            module.scope_acq_sequencer_select(scope_acq.sequencer_select)
        if scope_acq.trigger_mode_path0:
            module.scope_acq_trigger_mode_path0(scope_acq.trigger_mode_path0)
        if scope_acq.avg_mode_en_path0:
            module.scope_acq_avg_mode_en_path0(scope_acq.avg_mode_en_path0)
        if scope_acq.trigger_mode_path1:
            module.scope_acq_trigger_mode_path1(scope_acq.trigger_mode_path1)
        if scope_acq.avg_mode_en_path1:
            module.scope_acq_avg_mode_en_path1(scope_acq.avg_mode_en_path1)


class QrmRfConfigHelper(QrmConfigHelper):
    def configure_module(self, module: Module):
        self.configure_lo(module)
        self.configure_attenuation(module)
        self.configure_offset(module)
        self.configure_scope_acq(module)

    def configure_sequencer(self, sequencer: Sequencer):
        self.configure_connection(sequencer)
        self.configure_nco(sequencer)
        self.configure_awg(sequencer)
        self.configure_mixer(sequencer)
        self.configure_acq(sequencer)

    def configure_lo(self, module: Module):
        if self.module_config.lo.out0_in0_en:
            module.out0_in0_lo_en(self.module_config.lo.out0_in0_en)  # Switch the LO on
        if self.module_config.lo.out0_in0_freq:
            module.out0_in0_lo_freq(self.module_config.lo.out0_in0_freq)

    def configure_attenuation(self, module: Module):
        if self.module_config.attenuation.out0:
            module.out0_att(self.module_config.attenuation.out0)
        if self.module_config.attenuation.in0:
            module.in0_att(self.module_config.attenuation.in0)

    def configure_offset(self, module: Module):
        if self.module_config.offset.out0_path0:
            module.out0_offset_path0(self.module_config.offset.out0_path0)
        if self.module_config.offset.out0_path1:
            module.out0_offset_path1(self.module_config.offset.out0_path1)
        if self.module_config.offset.in0_path0:
            module.in0_offset_path0(self.module_config.offset.in0_path0)
        if self.module_config.offset.in0_path1:
            module.in0_offset_path1(self.module_config.offset.in0_path1)

    def configure_acq(self, sequencer: Sequencer):
        # Square weight integration
        if self.sequencer_config.demod_en_acq:
            sequencer.demod_en_acq(self.sequencer_config.demod_en_acq)
        if self.sequencer_config.square_weight_acq.integration_length:
            sequencer.integration_length_acq(
                self.sequencer_config.square_weight_acq.integration_length
            )
        if self.sequencer_config.thresholded_acq.rotation:
            sequencer.thresholded_acq_rotation(
                self.sequencer_config.thresholded_acq.rotation
            )
        if self.sequencer_config.thresholded_acq.threshold:
            sequencer.thresholded_acq_threshold(
                self.sequencer_config.thresholded_acq.threshold
            )
        if self.sequencer_config.thresholded_acq.trigger_en:
            sequencer.thresholded_acq_trigger_en(
                self.sequencer_config.thresholded_acq.trigger_en
            )
        if self.sequencer_config.thresholded_acq.trigger_address:
            sequencer.thresholded_acq_trigger_address(
                self.sequencer_config.thresholded_acq.trigger_address
            )
        if self.sequencer_config.thresholded_acq.trigger_invert:
            sequencer.thresholded_acq_trigger_invert(
                self.sequencer_config.thresholded_acq.trigger_invert
            )

    def calibrate_lo_leakage(self, module: Module):
        module.out0_in0_lo_cal()

        return OffsetConfig(
            out0_path0=module.out0_offset_path0(),
            out0_path1=module.out0_offset_path1(),
            in0_path0=module.in0_offset_path0(),
            in0_path1=module.in0_offset_path1(),
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from qat.backend.fallthrough import FallthroughBackend
from qat.frontend import AutoFrontend
from qat.middleend import CustomMiddleend
from qat.model.target_data import TargetData
from qat.pipelines.legacy.base import results_pipeline
from qat.pipelines.pipeline import Pipeline
from qat.pipelines.updateable import PipelineConfig, UpdateablePipeline
from qat.purr.backends.echo import EchoEngine
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.purr.utils.logger import get_default_logger
from qat.runtime.legacy import LegacyRuntime

from .compile import middleend_pipeline

log = get_default_logger()


class LegacyEchoPipeline(UpdateablePipeline):
    """A pipeline that compiles programs using the legacy echo backend and executes them
    using the :class:`LegacyRuntime`.

    Implements a custom pipeline to make instructions suitable for the legacy echo engine,
    and cannot be configured with a custom engine.
    """

    @staticmethod
    def _build_pipeline(
        config: PipelineConfig,
        model: QuantumHardwareModel,
        target_data: TargetData | None = None,
        engine=None,
    ) -> Pipeline:
        if engine is not None:
            log.warning(
                "An engine was provided to the LegacyEchoPipeline, but it will be ignored. "
                "The legacy EchoEngine is used directly."
            )

        target_data = target_data if target_data is not None else TargetData.default()
        return Pipeline(
            name=config.name,
            model=model,
            target_data=target_data,
            frontend=AutoFrontend.default_for_legacy(model),
            middleend=CustomMiddleend(
                model,
                pipeline=middleend_pipeline(model=model, target_data=target_data),
            ),
            backend=FallthroughBackend(model),
            runtime=LegacyRuntime(
                engine=EchoEngine(model=model),
                results_pipeline=results_pipeline(model),
            ),
        )
